#!/usr/bin/env python
#
# Copyright (c) 2015 10X Genomics, Inc. All rights reserved.
#

from __future__ import absolute_import, print_function
import os
import errno
import shutil
import subprocess
import sys
import gzip
import _io as io  # this is necessary b/c this module is named 'io' ... :(
from six import ensure_binary, string_types, PY3
import lz4.frame as lz4
import bmkcc.tenkit.log_subprocess as tk_subproc
import bmkcc.cellranger.h5_constants as h5_constants

if PY3:
    from typing import AnyStr  # pylint: disable=unused-import


def fixpath(path):
    # type: (str) -> str
    return os.path.abspath(os.path.expandvars(os.path.expanduser(path)))


def get_input_path(oldpath, is_dir=False):
    if not isinstance(oldpath, string_types):
        sys.exit("'{}' is not a valid string type and is an invalid path".format(oldpath))
    path = fixpath(oldpath)
    if not os.path.exists(path):
        sys.exit("Input file does not exist: %s" % path)
    if not os.access(path, os.R_OK):
        sys.exit("Input file path {} does not have read permissions".format(path))
    if is_dir:
        if not os.path.isdir(path):
            sys.exit("Please provide a directory, not a file: %s" % path)
    else:
        if not os.path.isfile(path):
            sys.exit("Please provide a file, not a directory: %s" % path)
    return path


def get_input_paths(oldpaths):
    paths = []
    for oldpath in oldpaths:
        paths.append(get_input_path(oldpath))
    return paths


def get_output_path(oldpath):
    path = fixpath(oldpath)
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        sys.exit("Output directory does not exist: %s" % dirname)
    if not os.path.isdir(dirname):
        sys.exit("Please provide a directory, not a file: %s" % dirname)
    return path


def open_maybe_gzip(filename, mode="r"):
    # type: (AnyStr, str) -> io.IOBase
    # this _must_ be a bytes
    if not isinstance(filename, bytes):
        filename = (
            ensure_binary(filename)
            if isinstance(filename, string_types)
            else ensure_binary(str(filename))
        )
    if filename.endswith(h5_constants.GZIP_SUFFIX):
        raw = gzip.open(filename, mode, 2)
    elif filename.endswith(h5_constants.LZ4_SUFFIX):
        raw = lz4.open(filename, mode)
    else:
        return open(filename, mode)

    bufsize = 1024 * 1024  # 1MB of buffering
    if mode == "r":
        return io.TextIOWrapper(io.BufferedReader(raw, buffer_size=bufsize))
    elif mode == "w":
        return io.TextIOWrapper(io.BufferedWriter(raw, buffer_size=bufsize))
    elif mode == "rb":
        return io.BufferedReader(raw, buffer_size=bufsize)
    elif mode == "wb":
        return io.BufferedWriter(raw, buffer_size=bufsize)

    else:
        raise ValueError("Unsupported mode for compression: %s" % mode)


class CRCalledProcessError(Exception):
    def __init__(self, msg):
        super(CRCalledProcessError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


def run_command_safely(cmd, args):
    p = tk_subproc.Popen([cmd] + args, stderr=subprocess.PIPE)
    _, stderr_data = p.communicate()
    if p.returncode != 0:
        raise Exception("%s returned error code %d: %s" % (p, p.returncode, stderr_data))


def check_completed_process(p, cmd):
    """p   (Popen object): Subprocess
    cmd (str):          Command that was run
    """
    if p.returncode is None:
        raise CRCalledProcessError("Process did not finish: %s ." % cmd)
    elif p.returncode != 0:
        raise CRCalledProcessError("Process returned error code %d: %s ." % (p.returncode, cmd))


def mkdir(dst, allow_existing=False):
    """Create a directory. Optionally succeed if already exists.
    Useful because transient NFS server issues may induce double creation attempts."""
    if allow_existing:
        try:
            os.mkdir(dst)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(dst):
                pass
            else:
                raise
    else:
        os.mkdir(dst)


def makedirs(dst, allow_existing=False):
    """Create a directory recursively. Optionally succeed if already exists.
    Useful because transient NFS server issues may induce double creation attempts."""
    if allow_existing:
        try:
            os.makedirs(dst)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(dst):
                pass
            else:
                raise
    else:
        os.makedirs(dst)


def remove(f, allow_nonexisting=False):
    """Delete a file. Allow to optionally succeed if file doesn't exist.
    Useful because transient NFS server issues may induce double deletion attempts."""
    if allow_nonexisting:
        try:
            os.remove(f)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise
    else:
        os.remove(f)


def copy(src, dst):
    """ Safely copy a file. Not platform-independent """
    run_command_safely("cp", [src, dst])


def move(src, dst):
    """ Safely move a file. Not platform-independent """
    run_command_safely("mv", [src, dst])


def copytree(src, dst, allow_existing=False):
    """ Safely recursively copy a directory. Not platform-independent """
    makedirs(dst, allow_existing=allow_existing)

    for name in os.listdir(src):
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)

        if os.path.isdir(srcname):
            copytree(srcname, dstname)
        else:
            copy(srcname, dstname)


def hardlink_with_fallback(src, dst):
    """Hard-links src to dst, falling back to copy if it fails.

    If `src` is a directory, it will attempt to recursively hardlink.
    """
    if os.path.isdir(src):
        # recursively hardlink a path, fallback to copy if fail
        try:
            shutil.copytree(src, dst, copy_function=os.link)
        except:
            shutil.copytree(src, dst)
    else:
        # hardlink a file, fallback to copy if fail
        try:
            os.link(src, dst)
        except:
            shutil.copy(src, dst)


def concatenate_files(out_path, in_paths, mode=""):
    with open(out_path, "w" + mode) as out_file:
        for in_path in in_paths:
            with open(in_path, "r" + mode) as in_file:
                shutil.copyfileobj(in_file, out_file)


def concatenate_headered_files(out_path, in_paths, mode=""):
    """Concatenate files, taking the first line of the first file
    and skipping the first line for subsequent files.
    Asserts that all header lines are equal."""
    with open(out_path, "w" + mode) as out_file:
        if len(in_paths) > 0:
            # Write first file
            with open(in_paths[0], "r" + mode) as in_file:
                header = in_file.readline()
                out_file.write(header)
                shutil.copyfileobj(in_file, out_file)

        # Write remaining files
        for in_path in in_paths[1:]:
            with open(in_path, "r" + mode) as in_file:
                this_header = in_file.readline()
                assert this_header == header
                shutil.copyfileobj(in_file, out_file)


def write_empty_json(filename):
    with open(filename, "wb") as f:
        f.write(b"{}")


def touch(path):
    # type: (str) -> None
    """Create an empty file"""
    fh = open(path, "w")
    fh.close()



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import

import bmkcc.cellranger.library_constants as lib_constants
import bmkcc.cellranger.webshim.constants.shared as shared
import bmkcc.cellranger.vdj.utils as vdj_utils


def vdj_gene_pair_format(gene_pair_str):
    """ Format a gene pair prefix from VdjReporter for human-usage """
    if gene_pair_str == lib_constants.MULTI_REFS_PREFIX:
        return ""
    else:
        return "(%s)" % ", ".join(vdj_utils.get_genes_in_pair(gene_pair_str))


# NOTE: Q30 metrics are currently shared with count.
#       However, RNA Read is renamed to RNA Read 1 here.
TOTAL_READ_PAIRS_METRIC = {
    "name": "VDJ_total_read_pairs",
    "display_name": "Number of Read Pairs",
    "description": "Total number of read pairs that were assigned to this library in demultiplexing.",
    "format": "integer",
}

Q30_METRICS = [
    {
        "name": "VDJ_bc_bases_with_q30_frac",
        "display_name": "Q30 Bases in Barcode",
        "description": "Fraction of cell barcode bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "VDJ_read_bases_with_q30_frac",
        "display_name": "Q30 Bases in RNA Read 1",
        "description": "Fraction of RNA read 1 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "VDJ_read2_bases_with_q30_frac",
        "display_name": "Q30 Bases in RNA Read 2",
        "description": "Fraction of RNA read 2 bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
    {
        "name": "VDJ_umi_bases_with_q30_frac",
        "display_name": "Q30 Bases in UMI",
        "description": "Fraction of UMI bases with Q-score >= 30, excluding very low quality/no-call (Q <= 2) bases from the denominator.",
        "format": "percent",
    },
]

GOOD_BCS_METRIC = {
    "name": "vdj_good_bc_frac",
    "display_name": "Valid Barcodes",
    "description": "Fraction of reads with barcodes that match the whitelist after barcode correction.",
    "format": "percent",
}

NUMBER_OF_CELLS_METRIC = {
    "name": "vdj_filtered_bcs",
    "display_name": "Estimated Number of Cells",
    "description": "The number of barcodes estimated to be associated with cells that express targeted V(D)J transcripts. A barcode is cell-associated if has 2 UMIs with sufficient read support.",
    "format": "integer",
}

MEAN_READ_PAIRS_PER_CELL_METRIC = {
    "name": "vdj_total_raw_read_pairs_per_filtered_bc",
    "display_name": "Mean Read Pairs per Cell",
    "description": "Number of input read pairs divided by the estimated number of cells.",
    "format": "integer",
}

MEAN_ASSEMBLED_READ_PAIRS_PER_CELL_METRIC = {
    "name": "vdj_assemblable_read_pairs_per_filtered_bc",
    "display_name": "Mean Used Read Pairs per Cell",
    "description": "Mean number of read pairs used in assembly per cell-associated barcode. These reads must have a cell-associated barcode, map to a V(D)J gene, and have a UMI with sufficient read support.",
    "format": "integer",
}

FRAC_READS_IN_CELLS_METRIC = {
    "name": "vdj_filtered_bcs_cum_frac",
    "display_name": "Fraction Reads in Cells",
    "description": "Number of reads with cell-associated barcodes divided by the number of reads with valid barcodes.",
    "format": "percent",
}

MULTI_READS_MAPPED_TO_RECOMBINOME_METRIC = {
    "name": "multi_vdj_recombinome_mapped_reads_frac",
    "display_name": "Reads Mapped to Any V(D)J Gene",
    "description": "Fraction of reads that partially or wholly map to any germline V(D)J gene segment.",
    "format": "percent",
}
READS_MAPPED_TO_RECOMBINOME_METRIC = {
    "name": "vdj_recombinome_mapped_reads_frac",
    "display_name": "Reads Mapped to %s",
    "description": "Fraction of reads that map partially or wholly to a germline %s gene segment.",
    "format": "percent",
    "prefix": "canonical_vdj_genes_nomulti",
}

MEDIAN_UMIS_PER_CELL_METRIC = {
    "name": "vdj_assembly_umis_per_cell_median",
    "display_name": "Median %s UMIs per Cell",
    "description": "Median number of UMIs assigned to a %s contig per cell. For B cells, only the max of {IGK, IGL} are counted.",
    "prefix": "canonical_vdj_genes_nomulti",
    "format": "%0.1f",
}
TRA_MEDIAN_UMIS_PER_CELL_METRIC = dict(MEDIAN_UMIS_PER_CELL_METRIC, prefix_filter="TRA")
TRB_MEDIAN_UMIS_PER_CELL_METRIC = dict(MEDIAN_UMIS_PER_CELL_METRIC, prefix_filter="TRB")

MULTI_NUM_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC = {
    "name": "multi_vdj_assembly_contig_pair_productive_full_len_bc_count",
    "display_name": "Number of Cells With Productive V-J Spanning Pair",
    "description": "Number of cell barcodes for which at least 1 sequence was found for each gene in a receptor pair. Both sequences must have an annotated CDR3, must be predicted to be productive, and must span the beginning of the 5' end of the V region to at least 2bp from the 3' end of the J region.",
    "format": "integer",
}
NUM_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC = {
    "name": "vdj_assembly_contig_pair_productive_full_len_bc_count",
    "display_name": "Number of Cells With Productive V-J Spanning %s Pair",
    "description": "Number of cell barcodes for which at least 1 sequence was found for each gene in a receptor pair. Both sequences must have an annotated CDR3, must be predicted to be productive, and must span the beginning of the 5' end of the V region to at least 2bp from the 3' end of the J region.",
    "prefix": "canonical_vdj_gene_pairs",
    "prefix_format": vdj_gene_pair_format,
    "format": "integer",
}
# # Make this name/prefix combo static internal target purposes
# TR_NUM_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC = {
#     'name': vdj_constants.CANONICAL_TR_GENE_PAIRS[0] + '_vdj_assembly_contig_pair_productive_full_len_bc_count',
#     'display_name': 'Number of Cells With Productive V-J Spanning %s Pair' % vdj_gene_pair_format(vdj_constants.CANONICAL_TR_GENE_PAIRS[0]),
#     'description': 'Number of cell barcodes for which at least 1 sequence was found for each of TRA and TRB. Both sequences must have an annotated CDR3, must be predicted to be productive, and must span the beginning of the 5\' end of the V region to at least 2bp from the the 3\' end of the J region.',
#     'format': 'integer',
# }

PAIRED_CLONOTYPE_DIVERSITY_METRIC = {
    "name": "multi_raw_vdj_paired_clonotype_diversity",
    "display_name": "Paired Clonotype Diversity",
    "description": "Effective diversity of the paired clonotypes, computed as the Inverse Simpson Index of the clonotype frequencies. A value of 1 indicates a minimally diverse sample - only one distinct clonotype was detected. A value equal to the estimated number of cells indicates a maximally diverse sample.",
    "format": "%0.2f",
}

MULTI_FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC = {
    "name": "multi_vdj_assembly_contig_pair_productive_full_len_bc_frac",
    "display_name": "Cells With Productive V-J Spanning Pair",
    "description": "Fraction of cell-associated barcodes with at least one contig for each chain of the receptor pair satisfying the following: "
    + "the contig annotations span the 5' end of the V region to at least 2 bp from the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region.",
    "format": "percent",
}
FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC = {
    "name": "vdj_assembly_contig_pair_productive_full_len_bc_frac",
    "display_name": "Cells With Productive V-J Spanning %s Pair",
    "description": "Fraction of cell-associated barcodes with at least one contig for each chain of the receptor pair satisfying the following: "
    + "the contig annotations span the 5' end of the V region to at least 2 bp from the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region.",
    "format": "percent",
    "prefix": "canonical_vdj_gene_pairs_nomulti",
    "prefix_format": vdj_gene_pair_format,
}

FRAC_CELLS_VDJ_GENE_DETECTED_METRIC = {
    "name": "vdj_assembly_contig_bc_frac",
    "display_name": "Cells With %s Contig",
    "description": "Fraction of cell-associated barcodes with at least one %s contig annotated as a full or partial V(D)J gene.",
    "format": "percent",
    "prefix": "canonical_vdj_genes_nomulti",
}

FRAC_CELLS_CDR3_DETECTED_METRIC = {
    "name": "vdj_assembly_cdr_detected_bc_frac",
    "display_name": "Cells With CDR3-annotated %s Contig",
    "description": "Fraction of cell-associated barcodes with at least one %s contig where a CDR3 was detected.",
    "format": "percent",
    "prefix": "canonical_vdj_genes_nomulti",
}

FRAC_CELLS_VJ_SPANNING_METRIC = {
    "name": "vdj_assembly_contig_full_len_bc_frac",
    "display_name": "Cells With V-J Spanning %s Contig",
    "description": "Fraction of cell-associated barcodes with at least one contig spanning the 5' end of the V region to at least 2bp from the 3' end of the J region for %s.",
    "format": "percent",
    "prefix": "canonical_vdj_genes_nomulti",
}

FRAC_CELLS_PRODUCTIVE_METRIC = {
    "name": "vdj_assembly_prod_cdr_bc_frac",
    "display_name": "Cells With Productive %s Contig",
    "description": "Fraction of cell-associated barcodes with at least one contig that spans the 5' end of the V region to at least 2bp from the 3' end of the J region for %s, has a start codon in the expected part of the V sequence, has an in-frame CDR3, and has no stop codons in the aligned V-J region.",
    "format": "percent",
    "prefix": "canonical_vdj_genes_nomulti",
}

SUMMARY_METRICS = [
    NUMBER_OF_CELLS_METRIC,
    MEAN_READ_PAIRS_PER_CELL_METRIC,
    MULTI_NUM_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC,
]

SEQUENCING_METRICS = [TOTAL_READ_PAIRS_METRIC, GOOD_BCS_METRIC,] + Q30_METRICS

ENRICHMENT_METRICS = [
    MULTI_READS_MAPPED_TO_RECOMBINOME_METRIC,
    READS_MAPPED_TO_RECOMBINOME_METRIC,
]

CELL_METRICS = [
    NUMBER_OF_CELLS_METRIC,
    MEAN_READ_PAIRS_PER_CELL_METRIC,
    MEAN_ASSEMBLED_READ_PAIRS_PER_CELL_METRIC,
    FRAC_READS_IN_CELLS_METRIC,
]

VDJ_EXPRESSION_METRICS = [
    MEDIAN_UMIS_PER_CELL_METRIC,
]

VDJ_ANNOTATION_METRICS = [
    MULTI_FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC,
    FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC,
    PAIRED_CLONOTYPE_DIVERSITY_METRIC,
    FRAC_CELLS_VDJ_GENE_DETECTED_METRIC,
    FRAC_CELLS_CDR3_DETECTED_METRIC,
    FRAC_CELLS_VJ_SPANNING_METRIC,
    FRAC_CELLS_PRODUCTIVE_METRIC,
]

METRICS = [
    {"name": "Summary", "metrics": SUMMARY_METRICS,},
    {"name": "Sequencing", "metrics": SEQUENCING_METRICS,},
    {"name": "Enrichment", "metrics": ENRICHMENT_METRICS,},
    {"name": "Cells", "metrics": CELL_METRICS,},
    {"name": "V(D)J Expression", "metrics": VDJ_EXPRESSION_METRICS,},
    {"name": "V(D)J Annotation", "metrics": VDJ_ANNOTATION_METRICS,},
]

CHARTS = [
    {
        "layout": {
            "title": "Barcode Rank",
            "width": 470,
            "height": 313,
            "margin": {"l": 60, "r": 0, "t": 30, "b": 40},
            "hovermode": "closest",
            "xaxis": {"title": "Barcodes", "type": "log",},
            "yaxis": {"title": "UMI counts", "type": "log",},
        },
        "data": [],
        "config": shared.CHARTS_PLOTLY_FIXED_CONFIG,
        "function": "plot_vdj_barcode_rank",
        "name": "vdj_barcode_rank",
    },
    {
        "layout": {
            "title": "Top 10 Clonotype Frequencies",
            "width": shared.DEFAULT_CHART_WIDTH,
            "height": shared.DEFAULT_CHART_HEIGHT,
            "showlegend": False,
            "xaxis": {"title": "Clonotype",},
            "yaxis": {"title": "Fraction of Cells",},
        },
        "data": [{"type": "bar"}],
        "kwargs": {
            "metric_name": "raw_vdj_clonotype_prop",
            "order_by": shared.HISTOGRAM_METRIC_ORDER_DECREASING_PROPORTION,
        },
        "config": shared.CHARTS_PLOTLY_FIXED_CONFIG,
        "function": "plot_histogram_metric",
        "name": "vdj_clonotype_histogram",
        "description": 'Top 10 clonotypes by frequency in this sample. A clonotype is defined as a unique set of CDR3 nucleotide sequences. For the full table and more details, please refer to the "clonotypes.csv" and "consensus_annotations.csv" files produced by the pipeline.',
    },
    {
        "table": {"width": 999, "height": 1000,},
        "title": "Top 10 Clonotype CDR3 Sequences",
        "name": "clonotype_table",
        "description": 'Top 10 clonotypes by frequency in this sample. A clonotype is defined as a unique set of CDR3 nucleotide sequences. For the full table and more details, please refer to the "clonotypes.csv" and "consensus_annotations.csv" files produced by the pipeline.',
        "function": "plot_clonotype_table",
    },
]

METRIC_ALARMS = [
    {
        "name": NUMBER_OF_CELLS_METRIC["name"],
        "format": NUMBER_OF_CELLS_METRIC["format"],
        "error": {
            "title": "No Cells Detected",
            "message": "No valid sequencing data was detected. Please check the sequencing data.",
            "test": "< 1",
        },
        "warn": {
            "title": "Low Number of Cells Detected",
            "message": "Ideal >= 10. This usually indicates poor cell quality, poor library quality, or poor sequencing quality. Application performance is likely to be affected.",
            "test": "< 10",
        },
    },
    {
        "name": MULTI_FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC["name"],
        "format": MULTI_FRAC_CELLS_PAIRED_VJ_SPANNING_PROD_METRIC["format"],
        "error": {
            "title": "Low Cells With Productive V-J Spanning Pair.",
            "message": "Ideal > 30%. This can indicate poor cell quality, low yield from the RT reaction, poor specificity of the V(D)J enrichment, poor sequencing quality, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance is likely to be affected",
            "test": "< 0.2",
        },
        "warn": {
            "title": "Low Cells With Productive V-J Spanning Pair.",
            "message": "Ideal > 30%. This can indicate poor cell quality, low yield from the RT reaction, poor specificity of the V(D)J enrichment, poor sequencing quality, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected",
            "test": "< 0.3",
        },
    },
    {
        "name": MULTI_READS_MAPPED_TO_RECOMBINOME_METRIC["name"],
        "format": MULTI_READS_MAPPED_TO_RECOMBINOME_METRIC["format"],
        "error": {
            "title": "Low Fraction of Reads Mapped to Any V(D)J Gene.",
            "message": "Ideal > 60%. This can indicate poor specificity of the V(D)J enrichment, use of the wrong germline reference, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance is likely to be affected.",
            "test": "< 0.50",
        },
        "warn": {
            "title": "Low Fraction of Reads Mapped to Any V(D)J Gene.",
            "message": "Ideal > 60%. This can indicate poor specificity of the V(D)J enrichment, use of the wrong germline reference, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.60",
        },
    },
    # Note: alerts don't support programmatic prefixes, so we have to hardcode TRA/TRB here.
    {
        "name": "TRA_vdj_assembly_umis_per_cell_median",
        "format": "%0.1f",
        "warn": {
            "title": "Zero Median TRA UMIs per Cell",
            "message": "Ideal > 0. This can indicate cells with extremely low TRA expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.1",
        },
        "filters": [{"chain_type": "TR"}],
    },
    {
        "name": "TRB_vdj_assembly_umis_per_cell_median",
        "format": "%0.1f",
        "warn": {
            "title": "Zero Median TRB UMIs per Cell",
            "message": "Ideal > 0. This can indicate cells with extremely low TRB expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.1",
        },
        "filters": [{"chain_type": "TR"}],
    },
    {
        "name": "IGH_vdj_assembly_umis_per_cell_median",
        "format": "%0.1f",
        "warn": {
            "title": "Zero Median IGH UMIs per Cell",
            "message": "Ideal > 0. This can indicate cells with extremely low IGH expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.1",
        },
        "filters": [{"chain_type": "IG"}],
    },
    {
        "name": "IGK_vdj_assembly_umis_per_cell_median",
        "format": "%0.1f",
        "warn": {
            "title": "Zero Median IGK UMIs per Cell",
            "message": "Ideal > 0. This can indicate cells with extremely low IGK expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.1",
        },
        "filters": [{"chain_type": "IG"}],
    },
    {
        "name": "IGL_vdj_assembly_umis_per_cell_median",
        "format": "%0.1f",
        "warn": {
            "title": "Zero Median IGL UMIs per Cell",
            "message": "Ideal > 0. This can indicate cells with extremely low IGL expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3' for V(D)J assembly). Application performance may be affected.",
            "test": "< 0.1",
        },
        "filters": [{"chain_type": "IG"}],
    },
    {
        "name": GOOD_BCS_METRIC["name"],
        "format": GOOD_BCS_METRIC["format"],
        "error": {
            "title": "Low Fraction Valid Barcodes",
            "message": "Ideal > 85%. This usually indicates a quality issue with the Ilumina R1 read. Application performance is likely to be affected.",
            "test": "< 0.75",
        },
        "warn": {
            "title": "Low Fraction Valid Barcodes",
            "message": "Ideal > 85%. This usually indicates a quality issue with the Ilumina R1 read. Application performance may be affected.",
            "test": "< 0.85",
        },
    },
    {
        "name": "bc_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low Barcode Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 90%. Application performance is likely to be affected.",
            "test": "< 0.75",
        },
        "warn": {
            "title": "Low Barcode Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 90%. Application performance may be affected.",
            "test": "< 0.90",
        },
    },
    {
        "name": "read_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low RNA Read Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 85%. Application performance is likely to be affected.",
            "test": "< 0.50",
        },
        "warn": {
            "title": "Low RNA Read Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 85%. Application performance may be affected.",
            "test": "< 0.85",
        },
    },
    {
        "name": "read2_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low RNA Read 2 Q30 Fraction (Illumina R2 for Single Cell V(D)J)",
            "message": "Ideal > 75%. Application performance is likely to be affected.",
            "test": "< 0.50",
        },
        "warn": {
            "title": "Low RNA Read 2 Q30 Fraction (Illumina R2 for Single Cell V(D)J)",
            "message": "Ideal > 75%. Application performance may be affected.",
            "test": "< 0.75",
        },
    },
    {
        "name": "umi_bases_with_q30_frac",
        "format": "percent",
        "error": {
            "title": "Low UMI Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 90%. Application performance is likely to be affected.",
            "test": "< 0.75",
        },
        "warn": {
            "title": "Low UMI Q30 Fraction (Illumina R1 for Single Cell V(D)J)",
            "message": "Ideal > 90%. Application performance may be affected.",
            "test": "< 0.90",
        },
    },
]



#!/usr/bin/env python
#
# Copyright (c) 2020 10X Genomics, Inc. All rights reserved.
#

# SINGLE CELL
def tc_metrics(is_spatial):
    if is_spatial:
        CELL_OR_TISSUE = "tissue"
        CELL_OR_SPOT = "Spot"
        TAS_OR_CELL = "tissue-associated spot"
        TAS_OR_CCP = "tissue-covered spot"
    else:
        CELL_OR_TISSUE = "cell"
        CELL_OR_SPOT = "Cell"
        TAS_OR_CELL = "cell"
        TAS_OR_CCP = "cell-containing partitions"

    NUM_GENES_ENRICHED_METRIC = {
        "name": "num_targeted_genes_enriched",
        "display_name": "Number of Targeted Genes Enriched Relative to Parent",
        "description": "Number of Targeted Genes Enriched. Only genes that have at least 1 "
        + CELL_OR_TISSUE
        + "-associated UMI in the parent sample are considered. Genes are classified with a two-class Gaussian mixture model according to their read enrichment relative to the parent sample. Genes in the group with the higher read enrichment are classified as enriched.",
        "format": "integer",
    }

    CELL_RELATIVE_TARGETED_DEPTH_METRIC = {
        "name": "cell_targeted_depth_factor",
        "display_name": "Mean Targeted Reads per " + CELL_OR_SPOT + " Relative to Parent",
        "description": "Mean ratio of targeted reads per  "
        + TAS_OR_CELL
        + " in the targeted sample relative to parent sample for all barcodes called as a "
        + TAS_OR_CELL
        + " in both samples.",
        "format": "%.2f",
    }

    TARGETED_ENRICHMENT_METRICS = [
        NUM_GENES_ENRICHED_METRIC,
        CELL_RELATIVE_TARGETED_DEPTH_METRIC,
    ]

    TOTAL_READ_PAIRS_TARGETED_METRIC = {
        "name": "total_read_pairs_in_targeted_sample",
        "display_name": "Number of Reads in Targeted Sample",
        "description": "Total number of read pairs in the targeted sample that were assigned to gene expression libraries in demultiplexing.",
        "format": "integer",
    }

    TOTAL_READ_PAIRS_PARENT_METRIC = {
        "name": "total_read_pairs_in_parent_sample",
        "display_name": "Number of Reads in Parent Sample",
        "description": "Total number of read pairs in the parent sample that were assigned to gene expression libraries in demultiplexing.",
        "format": "integer",
    }

    FRAC_ON_TARGET_TRANSCRIPTOME_TARGETED_METRIC = {
        "name": "fraction_reads_on_target_in_targeted_sample",
        "display_name": "Reads Confidently Mapped to the Targeted Transcriptome in Targeted Sample",
        "description": "Fraction of reads that mapped to a unique and targeted gene in the transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
        "format": "percent",
    }

    FRAC_ON_TARGET_TRANSCRIPTOME_PARENT_METRIC = {
        "name": "fraction_reads_on_target_in_parent_sample",
        "display_name": "Reads Confidently Mapped to the Targeted Transcriptome in Parent Sample",
        "description": "Fraction of reads that mapped to a unique and targeted gene in the transcriptome. The read must be consistent with annotated splice junctions. These reads are considered for UMI counting.",
        "format": "percent",
    }

    NUM_CELLS_TARGETED_METRIC = {
        "name": "num_cells_in_targeted_sample",
        "display_name": "Number of  " + TAS_OR_CELL.title() + "s Called in Targeted Sample",
        "description": "Number of barcodes called as  " + TAS_OR_CELL + "s in the Targeted Sample.",
        "format": "integer",
    }

    NUM_CELLS_PARENT_METRIC = {
        "name": "num_cells_in_parent_sample",
        "display_name": "Number of " + TAS_OR_CELL.title() + "s Called in Parent Sample",
        "description": "Number of barcodes called as " + TAS_OR_CELL + "s in the Parent Sample.",
        "format": "integer",
    }

    MEAN_RPC_TARGETED_METRIC = {
        "name": "mean_reads_per_cell_in_targeted_sample",
        "display_name": "Mean Reads per " + CELL_OR_SPOT + " in Targeted Sample",
        "description": "The total number of reads divided by the number of barcodes associated with "
        + TAS_OR_CCP
        + "s in the targeted sample.",
        "format": "integer",
    }

    MEAN_RPC_PARENT_METRIC = {
        "name": "mean_reads_per_cell_in_parent_sample",
        "display_name": "Mean Reads per " + CELL_OR_SPOT + " in Parent Sample",
        "description": "The total number of reads divided by the number of barcodes associated with "
        + TAS_OR_CCP
        + " in the parent sample.",
        "format": "integer",
    }

    MEAN_PRPC_TARGETED_METRIC = {
        "name": "mean_targeted_reads_per_cell_in_targeted_sample",
        "display_name": "Mean Targeted Reads per " + CELL_OR_SPOT + " in Targeted Sample",
        "description": "The total number of targeted reads divided by the number of barcodes associated with "
        + TAS_OR_CCP
        + "s in the targeted sample.",
        "format": "integer",
    }

    MEAN_PRPC_PARENT_METRIC = {
        "name": "mean_targeted_reads_per_cell_in_parent_sample",
        "display_name": "Mean Targeted Reads per " + CELL_OR_SPOT + " in Parent Sample",
        "description": "The total number of targeted reads divided by the number of barcodes associated with "
        + TAS_OR_CCP
        + "s in the parent sample.",
        "format": "integer",
    }

    MEDIAN_TARGETED_GENES_PER_CELL_TARGETED_METRIC = {
        "name": "median_targeted_genes_per_cell_in_targeted_sample",
        "display_name": "Median Targeted Genes per " + CELL_OR_SPOT + " in Targeted Sample",
        "description": "The median number of targeted genes detected per "
        + TAS_OR_CELL
        + " barcode in the targeted sample.",
        "format": "integer",
    }

    MEDIAN_TARGETED_GENES_PER_CELL_PARENT_METRIC = {
        "name": "median_targeted_genes_per_cell_in_parent_sample",
        "display_name": "Median Targeted Genes per " + CELL_OR_SPOT + " in Parent Sample",
        "description": "The median number of targeted genes detected per "
        + TAS_OR_CELL
        + "  barcode in the parent sample.",
        "format": "integer",
    }

    NUM_TARGETED_GENES_DETECTED_TARGETED_METRIC = {
        "name": "total_targeted_genes_detected_in_targeted_sample",
        "display_name": "Number of Targeted Genes Detected in Targeted Sample",
        "description": "Number of Targeted Genes Detected in Targeted Sample. A gene is considered detected if it has at least 1 "
        + CELL_OR_TISSUE
        + "-associated UMI.",
        "format": "integer",
    }

    NUM_TARGETED_GENES_DETECTED_PARENT_METRIC = {
        "name": "total_targeted_genes_detected_in_parent_sample",
        "display_name": "Number of Targeted Genes Detected in Parent Sample",
        "description": "Number of Targeted Genes Detected in Parent Sample. A gene is considered detected if it has at least 1 "
        + CELL_OR_TISSUE
        + "-associated UMI.",
        "format": "integer",
    }

    NUM_TARGETED_GENES_EXCLUSIVE_TARGETED_METRIC = {
        "name": "num_targeted_genes_detected_exclusive_in_targeted_sample",
        "display_name": "Number of Targeted Genes Detected Exclusively in Targeted Sample",
        "description": "Number of Targeted Genes Detected Exclusively in Targeted Sample. A gene is considered detected if it has at least 1 UMI count in "
        + CELL_OR_TISSUE
        + "-associated barcodes.",
        "format": "integer",
    }

    NUM_TARGETED_GENES_EXCLUSIVE_PARENT_METRIC = {
        "name": "num_targeted_genes_detected_exclusive_in_parent_sample",
        "display_name": "Number of Targeted Genes Detected Exclusively in Parent Sample",
        "description": "Number of Targeted Genes Detected Exclusively in Parent Sample. A gene is considered detected if it has at least 1 UMI count in "
        + CELL_OR_TISSUE
        + "-associated barcodes.",
        "format": "integer",
    }

    MEDIAN_TARGETED_UMIS_PER_CELL_TARGETED_METRIC = {
        "name": "median_targeted_umis_per_cell_in_targeted_sample",
        "display_name": "Median Targeted UMIs per " + CELL_OR_SPOT + " in Targeted Sample",
        "description": "The median number of UMI counts in targeted genes per "
        + CELL_OR_TISSUE
        + "-associated barcode in the targeted sample.",
        "format": "integer",
    }

    MEDIAN_TARGETED_UMIS_PER_CELL_PARENT_METRIC = {
        "name": "median_targeted_umis_per_cell_in_parent_sample",
        "display_name": "Median Targeted UMIs per " + CELL_OR_SPOT + " in Parent Sample",
        "description": "The median number of UMI counts in targeted genes per "
        + CELL_OR_TISSUE
        + "-associated barcode in the parent sample.",
        "format": "integer",
    }

    PAIRED_METRICS = [
        TOTAL_READ_PAIRS_TARGETED_METRIC,
        TOTAL_READ_PAIRS_PARENT_METRIC,
        FRAC_ON_TARGET_TRANSCRIPTOME_TARGETED_METRIC,
        FRAC_ON_TARGET_TRANSCRIPTOME_PARENT_METRIC,
        NUM_CELLS_TARGETED_METRIC,
        NUM_CELLS_PARENT_METRIC,
        MEAN_RPC_TARGETED_METRIC,
        MEAN_RPC_PARENT_METRIC,
        MEAN_PRPC_TARGETED_METRIC,
        MEAN_PRPC_PARENT_METRIC,
        MEDIAN_TARGETED_GENES_PER_CELL_TARGETED_METRIC,
        MEDIAN_TARGETED_GENES_PER_CELL_PARENT_METRIC,
        NUM_TARGETED_GENES_DETECTED_TARGETED_METRIC,
        NUM_TARGETED_GENES_DETECTED_PARENT_METRIC,
        NUM_TARGETED_GENES_EXCLUSIVE_TARGETED_METRIC,
        NUM_TARGETED_GENES_EXCLUSIVE_PARENT_METRIC,
        MEDIAN_TARGETED_UMIS_PER_CELL_TARGETED_METRIC,
        MEDIAN_TARGETED_UMIS_PER_CELL_PARENT_METRIC,
    ]

    NUM_CELLS_BOTH = {
        "name": "num_cell_calls_both",
        "display_name": "Number of " + CELL_OR_SPOT + "s Called in Parent and Targeted Samples",
        "description": "Number of barcodes called as "
        + TAS_OR_CELL
        + "s in both the parent and targeted samples.",
        "format": "integer",
    }

    NUM_CELLS_BOTH_TARGETED_ONLY = {
        "name": "num_cell_calls_targeted-only",
        "display_name": "Number of " + TAS_OR_CELL.title() + "s Called only in Targeted Sample",
        "description": "Number of barcodes called as "
        + TAS_OR_CELL
        + "s in the Targeted Sample and not the Parent Sample.",
        "format": "integer",
    }

    NUM_CELLS_PARENT_ONLY = {
        "name": "num_cell_calls_parent-only",
        "display_name": "Number of " + TAS_OR_CELL.title() + "s Called only in Parent Sample",
        "description": "Number of barcodes called as "
        + TAS_OR_CELL
        + "s in the Parent Sample and not the Targeted Sample.",
        "format": "integer",
    }

    MEAN_READ_ENRICHMENT = {
        "name": "mean_read_enrichment",
        "display_name": "Mean Read Enrichment across Targeted Genes",
        "description": "Mean Read Enrichment across Targeted Genes. Enrichments are only computed for genes with at least 1 "
        + CELL_OR_TISSUE
        + "-associated UMI in both the targeted and parent samples. Mean is the geometric mean of the enrichments. Samples are rescaled to the total number of reads in order to account for differences in sequencing depth.",
        "format": "%.3f",
    }

    PER_GENE_READ_CORR = {
        "name": "targeted_gene_read_rsquared",
        "display_name": "Per-Gene Read Counts R-squared",
        "description": "Pearson correlation coefficient (squared) of the number of reads confidently mapped to targeted genes (log10) in the targeted vs parent experiments.",
        "format": "%.3f",
    }

    MEAN_UMI_RATIO = {
        "name": "mean_frac_UMIs_recovered_per_gene",
        "display_name": "Mean Ratio of Targeted UMI Counts in Targeted Sample Relative to Parent",
        "description": "Mean per-gene ratio of UMIs observed in targeted sample relative to the parent sample.",
        "format": "%.3f",
    }

    PER_GENE_UMI_CORR = {
        "name": "targeted_gene_umi_rsquared",
        "display_name": "Per-Gene UMI Counts R-squared",
        "description": "Pearson correlation coefficient (squared) of the number of UMIs confidently mapped to targeted genes (log10) in the targeted vs parent experiments.",
        "format": "%.3f",
    }

    PLOTTED_METRICS = [
        NUM_CELLS_BOTH,
        NUM_CELLS_BOTH_TARGETED_ONLY,
        NUM_CELLS_PARENT_ONLY,
        MEAN_READ_ENRICHMENT,
        PER_GENE_READ_CORR,
        MEAN_UMI_RATIO,
        PER_GENE_UMI_CORR,
    ]

    METRICS = [
        {"name": "Targeted Enrichment", "metrics": TARGETED_ENRICHMENT_METRICS},
        {"name": "Paired Metrics", "metrics": PAIRED_METRICS},
        {"name": "Plotted Metrics", "metrics": PLOTTED_METRICS},
    ]

    return METRICS
    # ALARMS = []



#
# Copyright (c) 2019 10X Genomics, Inc. All rights reserved.
#

"""
Functions to compute various statistical operations on sparse matrices.
"""

from __future__ import absolute_import, division

import numpy as np
import scipy.stats
import sklearn.utils.sparsefuncs as sparsefuncs


def normalize_by_umi(matrix):
    counts_per_bc = matrix.get_counts_per_bc()
    median_counts_per_bc = max(1.0, np.median(counts_per_bc))
    scaling_factors = median_counts_per_bc / counts_per_bc

    # Normalize each barcode's total count by median total count
    m = matrix.m.copy().astype(np.float64)
    sparsefuncs.inplace_column_scale(m, scaling_factors)

    return m


def normalize_by_idf(matrix):
    """ Perform feature normalization."""
    numbcs_per_feature = matrix.get_numbcs_per_feature()
    scaling_factors_row = np.log(matrix.bcs_dim + 1) - np.log(1 + numbcs_per_feature)

    m = matrix.m.copy().astype(np.float64)
    sparsefuncs.inplace_row_scale(m, scaling_factors_row)

    # Extremely Rare Case (1 out of 1000s of samples tested):
    # Either the scaling or the count may be zero for all features for some barcode
    # This would lead to zero-ing out entire barcode upon normalization, which leads to a null
    # projection as well. This is harmful to analysis code that depends on at least a non-zero norm
    # for each barcode (e.g. spherical clustering and normalized tsne). We sprinkle in a small
    # value that ensures an nnz for the all-zero barcode, after finding such barcodes.

    # find zeroed barcodes and assign nnz to first feature (these barcodes are indistinguishable
    # anyway). We run the very small risk of making it similar to another barcode that is also nnz
    # in the first feature only
    zeroed = np.where(np.squeeze(np.asarray(m.sum(axis=0))) == 0)
    for bc_ix in zeroed:
        m[0, bc_ix] = 1e-15

    return m


def summarize_columns(matrix):
    """ Calculate mean and variance of each column, in a sparsity-preserving way."""
    mu, var = sparsefuncs.mean_variance_axis(matrix, axis=0)
    return np.array([mu]), np.array([var])


def get_normalized_dispersion(mat_mean, mat_var, nbins=20):
    """Calculates the normalized dispersion.  The dispersion is calculated for each feature
    and then normalized to see how its dispersion compares to samples that had a
    similar mean value.
    """
    # See equation in https://academic.oup.com/nar/article/40/10/4288/2411520
    # If a negative binomial is parameterized with mean m, and variance = m + d * m^2
    # then this d = dispersion as calculated below
    mat_disp = (mat_var - mat_mean) / np.square(mat_mean)

    quantiles = np.percentile(mat_mean, np.arange(0, 100, 100 // nbins))
    quantiles = np.append(quantiles, mat_mean.max())

    # merge bins with no difference in value
    quantiles = np.unique(quantiles)

    if len(quantiles) <= 1:
        # pathological case: the means are all identical. just return raw dispersion.
        return mat_disp

    # calc median dispersion per bin
    (disp_meds, _, disp_bins) = scipy.stats.binned_statistic(
        mat_mean, mat_disp, statistic="median", bins=quantiles
    )

    # calc median absolute deviation of dispersion per bin
    disp_meds_arr = disp_meds[disp_bins - 1]  # 0th bin is empty since our quantiles start from 0
    disp_abs_dev = abs(mat_disp - disp_meds_arr)
    (disp_mads, _, disp_bins) = scipy.stats.binned_statistic(
        mat_mean, disp_abs_dev, statistic="median", bins=quantiles
    )

    # calculate normalized dispersion
    disp_mads_arr = disp_mads[disp_bins - 1]
    disp_norm = (mat_disp - disp_meds_arr) / disp_mads_arr
    return disp_norm



#!/usr/bin/env python
#
# Copyright (c) 2014 10X Genomics, Inc. All rights reserved.
#
# General utilities for manipulating nucleotide sequences
#

from __future__ import absolute_import

import six

NUCS = [b"A", b"C", b"G", b"T"]
NUCS_INVERSE = {b"A": 0, b"C": 1, b"G": 2, b"T": 3}

if six.PY2:
    from string import maketrans  # pylint: disable=no-name-in-module
else:
    from typing import AnyStr  # pylint: disable=unused-import

    maketrans = bytes.maketrans  # pylint: disable=no-member

DNA_CONVERT_TABLE = maketrans(  # pylint: disable=deprecated-string-function
    b"ACGTacgtRYMKBDHVrymkbdhv", b"TGCAtgcaYRKMVHDByrkmvhdb"
)
RNA_CONVERT_TABLE = maketrans(  # pylint: disable=deprecated-string-function
    b"ACGUacguRYMKBDHVrymkbdhv", b"UGCAugcaYRKMVHDByrkmvhdb"
)

IUPAC_NUC_MAP = {
    b"A": [b"A"],
    b"C": [b"C"],
    b"G": [b"G"],
    b"T": [b"T"],
    b"R": [b"A", b"G"],
    b"Y": [b"C", b"T"],
    b"M": [b"C", b"A"],
    b"K": [b"T", b"G"],
    b"W": [b"T", b"A"],
    b"S": [b"C", b"G"],
    b"B": [b"C", b"T", b"G"],
    b"D": [b"T", b"A", b"G"],
    b"H": [b"T", b"A", b"C"],
    b"V": [b"A", b"C", b"G"],
    b"N": [b"T", b"A", b"C", b"G"],
}


def get_rev_comp(seq):
    # type: (AnyStr) -> bytes
    """Reverse complement for DNA.  Included ambiguous nucleotides and retains case."""
    return six.ensure_binary(seq).translate(DNA_CONVERT_TABLE)[::-1]


def mask(seq, keep_start, keep_end):
    # type: (bytes, int, int) -> bytes
    """Mask the sequence leaving only [keep_start, keep_end) unmasked"""
    return b"N" * keep_start + seq[keep_start:keep_end] + b"N" * (len(seq) - keep_end)



# Copyright (c) 2020 10x Genomics, Inc. All rights reserved.

""" A wrapper to avoid importing tables directly, we need to
monkey patch it on the fly to avoid issues with it changing thread usage
under the hood.  See:
https://10xtech.atlassian.net/projects/CELLRANGER/issues/CELLRANGER-2936
"""

from __future__ import absolute_import

import sys
import tables
import numexpr

# Hold if we've queried martian and the current threads to use
_CACHED_MARTIAN_STATUS = [False, numexpr.utils.detect_number_of_cores(), 0]


def _get_n_threads():
    """Memoized function to query martian for thread usage,
    assuming it's loaded.
    """
    if not _CACHED_MARTIAN_STATUS[0]:
        _CACHED_MARTIAN_STATUS[2] += 1
        try:
            import martian  # pylint: disable=bad-option-value,import-outside-toplevel

            n_threads = martian.get_threads_allocation()
            _CACHED_MARTIAN_STATUS[1] = n_threads
            tables.parameters.MAX_NUMEXPR_THREADS = n_threads
            _CACHED_MARTIAN_STATUS[0] = True
            sys.stdout.write(
                "Martian runtime detected. Try Num. = {}\n".format(_CACHED_MARTIAN_STATUS[2])
            )
        except:  # pylint: disable=bare-except
            sys.stdout.write(
                "Warning: No martian runtime detected. Try Num. = {}\n".format(
                    _CACHED_MARTIAN_STATUS[2]
                )
            )
    return _CACHED_MARTIAN_STATUS[1]


def file_constructor_wrapper(function):
    """A decorator around the initialization function
    to make sure it doesn't set the threads above what we want."""

    def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
        n_threads = _get_n_threads()
        kwargs["MAX_NUMEXPR_THREADS"] = n_threads
        kwargs["MAX_BLOSC_THREADS"] = n_threads
        return function(*args, **kwargs)

    return wrapper


# in case this value is used anywhere else

# specifically avoid any alternate settings by wrapping the constructor
tables.File.__init__ = file_constructor_wrapper(tables.File.__init__)



#!/usr/bin/env python
#
# Copyright (c) 2015 10X Genomics, Inc. All rights reserved.
#

from __future__ import absolute_import, division, print_function

import sys
import numpy as np
import scipy.stats as sp_stats
from scipy.special import loggamma
from six.moves import xrange

import bmkcc.tenkit.stats as tk_stats


def effective_diversity(counts):
    """ Inverse Simpson Index, or the effective diversity of power 2 """
    numerator = np.sum(counts) ** 2
    denominator = np.sum(v ** 2 for v in counts)
    return tk_stats.robust_divide(float(numerator), float(denominator))


def eval_multinomial_loglikelihoods(matrix, profile_p, max_mem_gb=0.1):
    """Compute the multinomial log PMF for many barcodes
    Args:
      matrix (scipy.sparse.csc_matrix): Matrix of UMI counts (feature x barcode)
      profile_p (np.ndarray(float)): Multinomial probability vector
      max_mem_gb (float): Try to bound memory usage.
    Returns:
      log_likelihoods (np.ndarray(float)): Log-likelihood for each barcode
    """
    gb_per_bc = float(matrix.shape[0] * matrix.dtype.itemsize) / (1024 ** 3)
    bcs_per_chunk = max(1, int(np.round(max_mem_gb / gb_per_bc)))
    num_bcs = matrix.shape[1]

    loglk = np.zeros(num_bcs)

    for chunk_start in xrange(0, num_bcs, bcs_per_chunk):
        chunk = slice(chunk_start, chunk_start + bcs_per_chunk)
        matrix_chunk = matrix[:, chunk].transpose().toarray()
        n = matrix_chunk.sum(1)
        loglk[chunk] = sp_stats.multinomial.logpmf(matrix_chunk, n, p=profile_p)
    return loglk


def simulate_multinomial_loglikelihoods(
    profile_p, umis_per_bc, num_sims=1000, jump=1000, n_sample_feature_block=1000000, verbose=False
):
    """Simulate draws from a multinomial distribution for various values of N.

       Uses the approximation from Lun et al. ( https://www.biorxiv.org/content/biorxiv/early/2018/04/04/234872.full.pdf )

    Args:
      profile_p (np.ndarray(float)): Probability of observing each feature.
      umis_per_bc (np.ndarray(int)): UMI counts per barcode (multinomial N).
      num_sims (int): Number of simulations per distinct N value.
      jump (int): Vectorize the sampling if the gap between two distinct Ns exceeds this.
      n_sample_feature_block (int): Vectorize this many feature samplings at a time.
    Returns:
      (distinct_ns (np.ndarray(int)), log_likelihoods (np.ndarray(float)):
      distinct_ns is an array containing the distinct N values that were simulated.
      log_likelihoods is a len(distinct_ns) x num_sims matrix containing the
        simulated log likelihoods.
    """
    np.random.seed(0)

    distinct_n = np.flatnonzero(np.bincount(umis_per_bc))

    loglk = np.zeros((len(distinct_n), num_sims), dtype=float)
    num_all_n = np.max(distinct_n) - np.min(distinct_n)
    if verbose:
        print("Number of distinct N supplied: %d" % len(distinct_n))
        print("Range of N: %d" % num_all_n)
        print("Number of features: %d" % len(profile_p))

    sampled_features = np.random.choice(
        len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True
    )
    k = 0

    log_profile_p = np.log(profile_p)

    for sim_idx in xrange(num_sims):
        if verbose and sim_idx % 100 == 99:
            sys.stdout.write(".")
            sys.stdout.flush()
        curr_counts = np.ravel(sp_stats.multinomial.rvs(distinct_n[0], profile_p, size=1))

        curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[0], p=profile_p)

        loglk[0, sim_idx] = curr_loglk

        for i in xrange(1, len(distinct_n)):
            step = distinct_n[i] - distinct_n[i - 1]
            if step >= jump:
                # Instead of iterating for each n, sample the intermediate ns all at once
                curr_counts += np.ravel(sp_stats.multinomial.rvs(step, profile_p, size=1))
                curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[i], p=profile_p)
                assert not np.isnan(curr_loglk)
            else:
                # Iteratively sample between the two distinct values of n
                for n in xrange(distinct_n[i - 1] + 1, distinct_n[i] + 1):
                    j = sampled_features[k]
                    k += 1
                    if k >= n_sample_feature_block:
                        # Amortize this operation
                        sampled_features = np.random.choice(
                            len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True
                        )
                        k = 0
                    curr_counts[j] += 1
                    curr_loglk += log_profile_p[j] + np.log(float(n) / curr_counts[j])

            loglk[i, sim_idx] = curr_loglk

    if verbose:
        sys.stdout.write("\n")

    return distinct_n, loglk


def compute_ambient_pvalues(umis_per_bc, obs_loglk, sim_n, sim_loglk):
    """Compute p-values for observed multinomial log-likelihoods
    Args:
      umis_per_bc (nd.array(int)): UMI counts per barcode
      obs_loglk (nd.array(float)): Observed log-likelihoods of each barcode deriving from an ambient profile
      sim_n (nd.array(int)): Multinomial N for simulated log-likelihoods
      sim_loglk (nd.array(float)): Simulated log-likelihoods of shape (len(sim_n), num_simulations)
    Returns:
      pvalues (nd.array(float)): p-values
    """
    assert len(umis_per_bc) == len(obs_loglk)
    assert sim_loglk.shape[0] == len(sim_n)

    # Find the index of the simulated N for each barcode
    sim_n_idx = np.searchsorted(sim_n, umis_per_bc)
    num_sims = sim_loglk.shape[1]

    num_barcodes = len(umis_per_bc)

    pvalues = np.zeros(num_barcodes)

    for i in xrange(num_barcodes):
        num_lower_loglk = np.sum(sim_loglk[sim_n_idx[i], :] < obs_loglk[i])
        pvalues[i] = float(1 + num_lower_loglk) / (1 + num_sims)
    return pvalues


class Curve(object):
    """Curve information for plotting

    Attributes
        x           (list of int): Curve x-axis (number of cells)
        y           (list of float): Curve y-axis (expected value of number of
                    unique clonotypes)
        y_std       (list of float): Standard deviation of number of unique
                    clonotypes in a curve
        y_ciu       (list of float): Upper bound of 95% confidence interval
                    for number of unique clonotypes in a curve
        y_cil       (list of float): Lower bound of 95% confidence interval
                    for number of unique clonotypes in a curve
    """

    def __init__(self):
        """Initialize an empty curve"""
        self.x = []
        self.y = []
        self.y_std = []
        self.y_ciu = []
        self.y_cil = []

    def is_empty(self):
        """Check if curve is empty or calculated

        Args:
            None
        Returns:
            bool
        """
        if len(self.y) == 0 or len(self.x) == 0:
            return True
        return False

    def is_consistent(self):
        """Check if calculated curve is consistent (length match)

        Args:
            None
        Returns:
            bool
        """
        len_list = [len(i) for i in [self.x, self.y, self.y_cil, self.y_ciu]]
        if len_list.count(len_list[0]) != len(len_list):
            return False
        return True


class Diversity(object):
    """Represents diversity with rarefaction and extrapolation curves.
    Attributes:
        sorted_hist         (list of tuple): A histogram sorted by abundance
                                            (most abundant has lowest index)
        freq_counts         (dict of int: int): Frequency counts table
        N                   (int): total number of counts in histogram
        rarefaction_curve   (Curve): Rarefaction curve information
        extrapolation_curve (Curve): Extrapolation curve information
    """

    def __init__(self, hist):
        self.sorted_hist = sorted(hist, reverse=True)
        self.N = self._calc_n()
        self.freq_counts = self._get_freq_counts()
        # Rarefaction curve
        self.rarefaction_curve = Curve()
        # Extrapolation curve
        self.extrapolation_curve = Curve()

    def is_diversity_curve_possible(self):
        if self.N == 0 or self.f_0_chao1() in [0, -1]:
            return False
        return True

    def _get_freq_counts(self):
        """Get frequency counts (f_k in Colwell et al 2012)

        Args:
            None

        Returns:
            dict
        """
        ret_dict = {}
        for abund in self.sorted_hist:
            if abund not in ret_dict:
                ret_dict[abund] = 1
            else:
                ret_dict[abund] += 1
        return ret_dict

    def _calc_n(self):
        """Calculates number of samples. (length of input histogram)

        Args:
            None

        Returns:
            int
        """
        return sum(self.sorted_hist)

    def _alpha(self, n, k):
        """Calculates alpha parameter in Colwell et al 2012

        Args:
            n (int): rarefaction point n
            k (int): Frequency

        Returns:
            int
        """
        if k > self.N - n:
            return 0
        log_alpha = (
            loggamma(self.N - k + 1)
            + loggamma(self.N - n + 1)
            - loggamma(self.N + 1)
            - loggamma(self.N - k - n + 1)
        )
        return np.exp(np.real(log_alpha))

    # Minimum variance unbiased estimator (MVUE) for both hypergeometric
    # and multinomial models
    # eq (4) in Colwell et al 2012
    # eq (5) in Colwell et al 2012
    def _rarefaction(self, n):
        """Calculates rarefaction and standard deviation of rarefaction for a single point n

        Args:
            n (int): rarefaction point n

        Returns:
            int, int
        """
        # number of observed clonotypes (S_obs in paper)
        num_clon = float(sum(self.freq_counts.values()))
        sum_helper_exp_val = 0
        for k in self.freq_counts:
            sum_helper_exp_val += self._alpha(n, k) * self.freq_counts[k]
        exp_val = num_clon - sum_helper_exp_val
        sum_helper_std_dev = 0
        for k in self.freq_counts:
            sum_helper_std_dev += (1 - self._alpha(n, k)) ** 2 * self.freq_counts[k] - float(
                exp_val ** 2
            ) / self.assemblage_size_estimate()
        return exp_val, np.sqrt(sum_helper_std_dev)

    def calc_rarefaction_curve(self, num_steps=40):
        """Calculates rarefaction curve for num_steps between 1 and N

        Args:
            num_steps (int): Number of steps between 1 and N

        Returns:
            None
        """
        step_size = int(self.N // (num_steps - 1))
        if step_size == 0:
            step_size = 1
        rc_x = list(range(1, self.N, step_size))
        if len(rc_x) < num_steps:
            rc_x = list(range(1, self.N + step_size, step_size))

        rc_y_exp = [None] * len(rc_x)
        rc_y_std = [None] * len(rc_x)
        rc_y_ciu = [None] * len(rc_x)
        rc_y_cil = [None] * len(rc_x)
        for (idx, n) in enumerate(rc_x):
            rc_y_exp[idx], rc_y_std[idx] = self._rarefaction(n)
            rc_y_cil[idx] = rc_y_exp[idx] - 1.96 * rc_y_std[idx]
            rc_y_ciu[idx] = rc_y_exp[idx] + 1.96 * rc_y_std[idx]
        self.rarefaction_curve.x = rc_x
        self.rarefaction_curve.y = rc_y_exp
        self.rarefaction_curve.y_std = rc_y_std
        self.rarefaction_curve.y_cil = rc_y_cil
        self.rarefaction_curve.y_ciu = rc_y_ciu

    def calc_rarefaction_curve_plotly(self, origin, color, num_steps=40, stdev=True):
        """Create a plotly curve for rarefaction.

        Args:
            origin (str): Origin string (see vdj inputs)
            color: color of the curve
            num_steps (int): number of extrapolation steps
            stdev (bool): flag for calculation of standard deviation (and confidence intarvals)


        Returns:
            [dict]
        """
        self.calc_rarefaction_curve(num_steps)
        lines = []
        lines.append(
            {
                "x": self.rarefaction_curve.x,
                "y": self.rarefaction_curve.y,
                "type": "scatter",
                "name": origin,
                "mode": "lines",
                "line_color": color,
            }
        )
        if stdev:
            lines.append(
                {
                    "x": self.rarefaction_curve.x + self.rarefaction_curve.x[::-1],
                    "y": self.rarefaction_curve.y_ciu + self.rarefaction_curve.y_cil[::-1],
                    "type": "scatter",
                    "name": origin + " (Stdev)",
                    "fill": "toself",
                    "line_color": "rgba(255,255,255,0)",
                    "fillcolor": color,
                    "opacity": 0.2,
                }
            )
        return lines

    # eq (9) in Colwell et al 2012
    # not using the approximation used in eq (9)
    def _extrapolation(self, n_plus_m):
        """Calculates extrapolation for a single point (N + m)
        standard deviation is not implemented.

        Args:
            n_plus_m (int): Distance of extrapolation point to N
                (how many more samples collected)

        Returns:
            int, int
        """
        # number of observed clonotypes (s_obs in paper)
        s_obs = float(sum(self.freq_counts.values()))
        f_0 = float(self.f_0_chao1())
        f_1 = float(self.freq_counts[1])

        N = float(self.N)  # n in paper
        m = float(n_plus_m - self.N)

        brackets = 1.0 - (1.0 - f_1 / N / f_0) ** m

        exp_val = s_obs + f_0 * brackets
        # Standard deviation for extrapolation is not implemented
        std_dev = None

        return exp_val, std_dev

    def calc_extrapolation_curve(self, num_steps=40, max_n=2000):
        """Calculates extrapolation curve for num_steps between N and max_n

        Args:
            num_steps (int): Number of steps between N and max_n
            max_n (int): The limit to which the function extrapolates

        Returns:
            None
        """
        step_size = int((max_n - self.N) / (num_steps - 1))
        if len(list(range(self.N, max_n, step_size))) < num_steps:
            ec_x = list(range(self.N, max_n + step_size, step_size))
        else:
            ec_x = list(range(self.N, max_n, step_size))
        ec_y_exp = [None] * len(ec_x)
        ec_y_std = [None] * len(ec_x)
        ec_y_ciu = [None] * len(ec_x)
        ec_y_cil = [None] * len(ec_x)
        for (idx, n_plus_m) in enumerate(ec_x):
            ec_y_exp[idx], ec_y_std[idx] = self._extrapolation(n_plus_m)
            if ec_y_std[idx] is not None:
                ec_y_cil[idx] = ec_y_exp[idx] - 1.96 * ec_y_std[idx]
                ec_y_ciu[idx] = ec_y_exp[idx] + 1.96 * ec_y_std[idx]
        self.extrapolation_curve.x = ec_x
        self.extrapolation_curve.y = ec_y_exp
        self.extrapolation_curve.y_std = ec_y_std
        self.extrapolation_curve.y_cil = ec_y_cil
        self.extrapolation_curve.y_ciu = ec_y_ciu

    def f_0_chao1(self):
        """Estimate f0 (number of clonotypes that we haven't seen yet) using chao1 estimator
        Based on: Colwell et al 2012

        Args:
            None

        Returns:
            float
        """
        if 1 in self.freq_counts and 2 in self.freq_counts:
            return float(self.freq_counts[1] ** 2) / float(2 * self.freq_counts[2])
        elif 1 in self.freq_counts:
            return float(self.freq_counts[1] * (self.freq_counts[1] - 1)) / 2.0
        else:
            # print "Error in f_0_chao1" # TODO log message
            return -1.0

    def assemblage_size_estimate(self):
        """Estimate assemblage size (number of total unique clonotypes) using chao1 estimator
        Based on: Colwell et al 2012

        Args:
            None

        Returns:
            float
        """
        return self.N + self.f_0_chao1()

    def create_both_curves(self, outfile, rc_steps=40, ec_steps=40, max_n=2000):
        """Calculate rarefaction and extrapolation diversity curves and save them
        in csv format in outfile.

        Args:
            outfile (str): Path to output file
            rc_steps (int): Rarefaction curve steps between 1 and N
            ec_steps (int): Extrapolation steps between N and max_n
            max_n (int): Maximum number of cells extrapolating to

        Returns:
            None
        """
        # Check if extrapolation_curve is calculated. If not, call the appropriate func to calc
        if self.extrapolation_curve.is_empty():
            self.calc_extrapolation_curve(ec_steps, max_n)
        # Check if rarefactoion_curve is calculated. If not, call the appropriate func to calc
        if self.rarefaction_curve.is_empty():
            self.calc_rarefaction_curve(rc_steps)
        if ~self.rarefaction_curve.is_consistent():
            return None  # TODO add appropriate logging message
        if ~self.extrapolation_curve.is_consistent():
            return None  # TODO add appropriate logging message
        header = "x,y,y_cil,y_ciu,type\n"
        with open(outfile, "w") as outhandle:
            outhandle.write(header)
            for (idx, x) in enumerate(self.rarefaction_curve.x):
                numbers = [
                    x,
                    self.rarefaction_curve.y[idx],
                    self.rarefaction_curve.y_cil[idx],
                    self.rarefaction_curve.y_ciu[idx],
                ]
                line = ",".join([str(n) for n in numbers] + ["rarefaction"]) + "\n"
                outhandle.write(line)
            for (idx, x) in enumerate(self.extrapolation_curve.x):
                numbers = [
                    x,
                    self.extrapolation_curve.y[idx],
                    self.extrapolation_curve.y_cil[idx],
                    self.extrapolation_curve.y_ciu[idx],
                ]
                line = ",".join([str(n) for n in numbers] + ["extrapolation"]) + "\n"
                outhandle.write(line)
        return 1



#!/usr/bin/env python
#
# Copyright (c) 2019 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import, division

import collections
import copy
import csv
import math
import random
import re
import sys
import numpy as np
from six import iteritems, itervalues, viewvalues, PY3
from six.moves import xrange as range, zip
import bmkcc.tenkit.safe_json as tk_safe_json
import bmkcc.cellranger.analysis.clustering as cr_clustering
#from cellranger.analysis.singlegenome import SingleGenomeAnalysis
#from cellranger.analysis.multigenome import MultiGenomeAnalysis
import bmkcc.cellranger.constants as cr_constants
#import cellranger.reference as cr_reference
#import cellranger.report as cr_report  # pylint: disable=no-name-in-module
import bmkcc.cellranger.rna.library as rna_library
import bmkcc.cellranger.utils as cr_utils
#import cellranger.targeted.utils as cr_tgt_utils
#import cellranger.vdj.constants as vdj_constants
#import cellranger.vdj.report as vdj_report
import bmkcc.cellranger.webshim.constants.shared as shared_constants
import bmkcc.cellranger.webshim.constants.gex as ws_gex_constants
#import cellranger.webshim.constants.vdj as ws_vdj_constants
from bmkcc.cellranger.webshim.data import SampleData, generate_counter_barcode_rank_plot_data
'''
from cellranger.websummary.sample_properties import (
    CountSampleProperties,
    SampleProperties,
)
'''
#from cellranger.websummary import react_components as rc
#from cellranger.analysis import jibes
#from cellranger.feature.jibes_tag_assigner import JibesTagAssigner
from six import ensure_str

if PY3:
    from typing import Any, Dict, List  # pylint: disable=unused-import

JIBES_BIPLOT_N_POINTS = 100000
MIN_QV_VAL = 1.0
MAX_QV_VAL = 100.0
# sample 100k points for jibes biplot

# https://yutannihilation.github.io/allYourFigureAreBelongToUs/ggthemes/economist_pal/
JIBES_COLOR_PALETTE = [
    "#3E647D",
    "#7B92AB",
    "#82C0E9",
    "#2D6D66",
    "#BFA19C",
    "#008BBC",
    "#97B6B0",
    "#D7D29E",
    "#1A476F",
    "#90353B",
    "#9C8847",
    "#938DD2",
    "#C10534",
    "#6E8E84",
    "#CAC27E",
]


def add_prefix(prefix, name):
    if "%s" in name:
        return name % prefix
    else:
        return "_".join([prefix, name]) if prefix else name


def format_name(display_name, prefix, prefixes, prefix_format_func=None):
    if prefix is not None and prefix_format_func is not None:
        prefix = prefix_format_func(prefix)

    # Default multi -> '' if no format func given
    if prefix_format_func is None and prefix == rna_library.MULTI_REFS_PREFIX:
        prefix = ""

    if len(prefixes) > 1 or "%s" in display_name:
        display_name = add_prefix(prefix, display_name)

    # Replace underscores w/ spaces
    display_name = display_name.replace("_", " ")

    # Collapse whitespace
    display_name = re.sub(r"\s+", " ", display_name)

    return display_name


def format_description(description, prefix, prefixes, prefix_format_func=None):
    if prefix is not None and prefix_format_func is not None:
        prefix = prefix_format_func(prefix)

    # Default multi -> '' if no format func given
    if prefix_format_func is None and prefix == rna_library.MULTI_REFS_PREFIX:
        prefix = ""

    if "%s" in description:
        # Escape stray percents
        description = re.sub("%([^s])", "%%\\1", description)

        # Only add the prefix if there are multiple possibilities
        s = str(prefix) if len(prefixes) > 1 and prefix else ""
        description = description % s

    # Collapse whitespace
    description = re.sub(r"\s+", " ", description)

    return description


def format_value(value, format_type):
    if format_type == "string":
        return str(value)

    value = value if not math.isnan(float(value)) else 0

    if format_type == "percent":
        return "{:.1%}".format(float(value))
    elif format_type == "integer":
        return "{:,.0f}".format(value)
    elif format_type[0] == "%":
        return format_type % float(value)
    raise Exception("Invalid format type: %s" % format_type)


def lookup_name(data, name):
    name_parts = name.split("/", 1)
    data = data.get(name_parts[0])
    if len(name_parts) == 1 or data is None:
        return data
    else:
        return lookup_name(data, name_parts[1])


def add_alarm(value, formatted_value, level, alarm_dict, alarms):
    # The numeric value may have become a string via JSON ('NaN', 'Infinity', '-Infinity')
    # Note that by default python's json encoder outputs a bare NaN symbol which is even worse
    test = 'float("%s") %s' % (str(value), alarm_dict["test"])
    raised = eval(test, {}, {})
    if raised:
        title = alarm_dict["title"]
        message = alarm_dict["message"]
        alarms[title] = {
            "title": title,
            "message": message,
            "value": formatted_value,
            "level": level,
        }
    return raised


def add_alarms(data, alarms_dict, alarms):
    # Prefixes on alarm metrics not implemented
    full_name = alarms_dict["name"]
    value = lookup_name(data, full_name)
    if value is None:
        return

    formatted_value = format_value(value, alarms_dict["format"])
    error_dict = alarms_dict.get(shared_constants.ALARM_ERROR)
    if error_dict is not None and add_alarm(
        value, formatted_value, shared_constants.ALARM_ERROR, error_dict, alarms
    ):
        return
    warn_dict = alarms_dict.get(shared_constants.ALARM_WARN)
    if warn_dict is not None:
        add_alarm(value, formatted_value, shared_constants.ALARM_WARN, warn_dict, alarms)


def add_table_rows(data, name, metric_dict, rows, style_func, target_func, prefixes=None):
    """Args:
    data (dict) - Dict containing summary metrics
    name (str) - Metric name (key in data dict)
    metric_dict (dict) - Metric display definition dict
    """
    format_type = metric_dict["format"]
    display_name = metric_dict.get("display_name", name)
    description = metric_dict.get("description", display_name)

    if not prefixes:
        prefixes = [None]

    values = []
    for prefix in prefixes:
        # Find the metric and construct a table row
        full_name = add_prefix(prefix, name)
        value = lookup_name(data, full_name)

        if value is not None:
            prefix_format_func = metric_dict.get("prefix_format")
            formatted_name = format_name(
                display_name, prefix, prefixes, prefix_format_func=prefix_format_func
            )
            formatted_description = format_description(
                description, prefix, prefixes, prefix_format_func=prefix_format_func
            )
            formatted_value = format_value(value, format_type)

            style = style_func(metric_dict, value)

            rows.append(
                [
                    {"v": formatted_name, "f": formatted_description, "s": style,},
                    {"v": target_func(metric_dict), "s": style,},
                    {"v": formatted_value, "s": style, "r": value,},
                ]
            )
            values.append(value)


def _is_metric_hidden(sample_properties, metric_dict):
    if "hidden" in metric_dict:
        return eval(metric_dict["hidden"], sample_properties, sample_properties)
    return False


def build_tables(
    sample_properties,
    table_dicts,
    alarm_table_dicts,
    sample_data,
    style_func=lambda *args: "",
    target_func=lambda *args: "",
    metric_headers=[],
    all_prefixes={},
):
    tables = []
    for table_dict in table_dicts:
        table_name = table_dict["name"]
        table_metrics = table_dict["metrics"]

        rows = []
        for metric_dict in table_metrics:
            name = metric_dict["name"]
            prefix = metric_dict.get("prefix")

            if _is_metric_hidden(sample_properties, metric_dict):
                continue

            if prefix is not None:
                # Search sample properties for prefix values
                prefixes = sample_properties.get(prefix)

                # Otherwise search the given prefix values
                if not prefixes:
                    prefixes = all_prefixes.get(prefix)

                # Apply the prefix filter for this metric
                if "prefix_filter" in metric_dict:
                    if prefixes is not None:
                        prefixes = [x for x in prefixes if x in metric_dict["prefix_filter"]]
                    else:
                        sys.stderr.write(
                            "Warning: no metric prefix values to filter for metric %s. The prefix name %s is probably wrong.\n"
                            % (name, prefix)
                        )

                add_table_rows(
                    sample_data.summary,
                    name,
                    metric_dict,
                    rows,
                    style_func,
                    target_func,
                    prefixes=prefixes,
                )
            else:
                add_table_rows(
                    sample_data.summary, name, metric_dict, rows, style_func, target_func
                )

        if len(rows) > 0:
            tables.append(
                {"name": table_name, "rows": rows, "headers": metric_headers,}
            )

        # TODO: Temporary hack to view antibody rank plot, until the new websummary arrives
        if (len(rows) == 0 and table_name == "Mapping") or (
            len(rows) == 0 and table_name == "Cells"
        ):
            tables.append(
                {"name": table_name, "rows": rows, "headers": metric_headers,}
            )

    # Build alarm table
    alarms = collections.OrderedDict()
    for alarm_dict in alarm_table_dicts:
        add_alarms(sample_data.summary, alarm_dict, alarms)

    return tables, list(viewvalues(alarms))


def convert_numpy_array_to_line_chart(array, ntype):
    """
    Given an array of values, sort them in descending order, then give
    the x,y coordinates for a line that could represent this sorted order by collapsing
    identical Y values into a single range of ranks.

    Example:
        array = np.array([1,1,1,2,3,3,3,3,3,4,5,9,20])
        convert_numpy_array_to_line_chart(array, float)
        # Result
        [[0, 20.0],
         [1, 9.0],
         [2, 5.0],
         [3, 4.0],
         [4, 3.0],
         [8, 3.0],
         [9, 2.0],
         [10, 1.0],
         [12, 1.0]]
    :param array: a numpy array
    :param ntype: a datatype to return the array values as
    :return:
    """
    array = np.sort(array)[::-1]
    rows = []
    previous_count = None
    for (index,), count in np.ndenumerate(array):
        if index == 0 or index == len(array) - 1:
            rows.append([index, ntype(count)])
        elif previous_count != count:
            assert previous_count is not None
            previous_index = rows[-1][0]
            if previous_index != index - 1:
                rows.append([index - 1, ntype(previous_count)])
            rows.append([index, ntype(count)])
        previous_count = count
    return rows


def _plot_barcode_rank(chart, counts, num_cells):
    """ Generate a generic barcode rank plot """
    rows = convert_numpy_array_to_line_chart(counts, int)

    for row in rows:
        index, count = row[0], row[1]
        if index < num_cells:
            series_list = [chart["data"][0]]
        elif index == num_cells:
            # Connect the two lines
            series_list = [chart["data"][0], chart["data"][1]]
        else:
            series_list = [chart["data"][1]]

        for series in series_list:
            series["x"].append(index)
            series["y"].append(count)

    # Handle case where there is no background
    bg_series = chart["data"][1]
    if len(bg_series["x"]) == 0:
        bg_series["x"].append(0)
        bg_series["y"].append(0)

    return chart


def build_plot_data_dict(plot_segment, counts, show_name_and_hover=True, color=None):
    """
    Construct the data for a plot segment by appropriately slicing the
    counts
    Inputs:
    - plot_segment: BarcodeRankPlotSegment containing [start, end)
        of the segment, the cell density and legend visibility option
    - counts: Reverse sorted UMI counts for all barcodes.
    - show_name_and_hover: boolean whether to add hover and name text specific to the classic
    CR rank plot (the user might want to set these themselves for alternate ranke plots)
    """

    # -1 for continuity between two charts
    start = max(0, plot_segment.start - 1)
    end = plot_segment.end
    plot_rows = convert_numpy_array_to_line_chart(counts[start:end], int)

    if color is None:
        color = shared_constants.BC_PLOT_CMAP(plot_segment.cell_density)

    data_dict = {
        "x": [],
        "y": [],
        "type": "scattergl",
        "mode": "lines",
        "line": {"color": color, "width": shared_constants.BC_RANK_PLOT_LINE_WIDTH,},
        "showlegend": plot_segment.legend,
    }
    if show_name_and_hover:
        name = "Cells" if plot_segment.cell_density > 0 else "Background"
        data_dict["name"] = name
        # Setup the tooltip
        if plot_segment.cell_density > 0.0:
            n_barcodes = plot_segment.end - plot_segment.start
            n_cells = int(
                round(plot_segment.cell_density * n_barcodes)  # pylint: disable=round-builtin
            )
            hover = "{:.0f}% Cells<br>({}/{})".format(
                100 * plot_segment.cell_density, n_cells, n_barcodes
            )
        else:
            hover = "Background"
        data_dict["hoverinfo"] = "text"
        data_dict["text"] = hover
    offset = 1 + start  # it's a log-log plot, hence the 1
    for index, count in plot_rows:
        data_dict["x"].append(index + offset)
        data_dict["y"].append(count)

    # Handle case where the data is empty
    if len(data_dict["x"]) == 0:
        data_dict["x"].append(0)
        data_dict["y"].append(0)

    return data_dict


def _plot_segmented_barcode_rank(chart, counts, plot_segments):
    """
    Generate the RNA counter barcode rank plot
    Inputs:
        - chart: chart element to populate data
        - counts: UMI counts reverse sorted
        - plot_segments: A list of BarcodeRankPlotSegments
    """

    for segment in plot_segments:
        chart["data"].append(build_plot_data_dict(segment, counts))

    return chart


def plot_basic_barcode_rank(chart, cell_barcodes, barcode_summary, genomes, lib_prefix):
    """
     Generate a basic RNA counter barcode rank plot without depending on SampleData/SampleProperties
    Inputs:
        - chart: chart element to populate data
        - cell_barcodes: set of cell barcodes as bytes
        - barcode_summary: barcode summary from the barcode_summary.h5
        - lib_prefix: The library prefix to create the plot for
    """
    genome = rna_library.MULTI_REFS_PREFIX

    if (
        lib_prefix
        == rna_library.get_library_type_metric_prefix(rna_library.GENE_EXPRESSION_LIBRARY_TYPE)
        and len(genomes) == 1
    ):
        genome = genomes[0]

    key = cr_utils.format_barcode_summary_h5_key(
        lib_prefix,
        genome,
        cr_constants.TRANSCRIPTOME_REGION,
        cr_constants.CONF_MAPPED_DEDUPED_READ_TYPE,
    )

    if key not in barcode_summary:
        return None

    counts_per_bc, plot_segments = generate_counter_barcode_rank_plot_data(
        cell_barcodes, barcode_summary, key
    )
    return _plot_segmented_barcode_rank(chart, counts_per_bc, plot_segments)


def plot_barcode_rank(chart, sample_properties, sample_data):
    """ Generate the RNA counter barcode rank plot """

    # TODO: The old PD plots use this function and pass things around as dictionaries,
    # so we convert here if this function was called from the PD code
    if isinstance(sample_properties, dict):
        sample_properties = CountSampleProperties(
            sample_id=sample_properties.get("sample_id", None),
            sample_desc=sample_properties.get("sample_desc", None),
            genomes=sample_properties.get("genomes", None),
        )

    assert isinstance(sample_data, SampleData)
    assert isinstance(sample_properties, SampleProperties)
    if (
        not isinstance(sample_properties, CountSampleProperties)
        or sample_data.barcode_summary is None
        or sample_data.cell_barcodes is None
    ):
        return None

    if len(sample_properties.genomes) == 0:
        return None

    # UMI counts per BC across all genomes present
    if len(sample_properties.genomes) > 1:
        genome = rna_library.MULTI_REFS_PREFIX
    else:
        genome = sample_properties.genomes[0]

    gex_prefix = rna_library.get_library_type_metric_prefix(
        rna_library.GENE_EXPRESSION_LIBRARY_TYPE
    )
    ab_prefix = rna_library.get_library_type_metric_prefix(rna_library.ANTIBODY_LIBRARY_TYPE)
    gex_key = cr_utils.format_barcode_summary_h5_key(
        gex_prefix,
        genome,
        cr_constants.TRANSCRIPTOME_REGION,
        cr_constants.CONF_MAPPED_DEDUPED_READ_TYPE,
    )
    ab_key = cr_utils.format_barcode_summary_h5_key(
        ab_prefix,
        rna_library.MULTI_REFS_PREFIX,
        cr_constants.TRANSCRIPTOME_REGION,
        cr_constants.CONF_MAPPED_DEDUPED_READ_TYPE,
    )

    if gex_key in sample_data.barcode_summary:
        counts_per_bc, plot_segments = sample_data.counter_barcode_rank_plot_data(gex_key)
        return _plot_segmented_barcode_rank(chart, counts_per_bc, plot_segments)
    elif ab_key in sample_data.barcode_summary:  # in case there was only Antibody library
        counts_per_bc, plot_segments = sample_data.counter_barcode_rank_plot_data(ab_key)
        return _plot_segmented_barcode_rank(chart, counts_per_bc, plot_segments)
    else:
        # Not guaranteed to exist, depending on pipeline
        return None


# TODO: This unneccesary argument is needed because of how the PD code is structured.
def plot_vdj_barcode_rank(chart, sample_properties, sample_data):
    """ Generate the VDJ barcode rank plot """
    if not sample_data.cell_barcodes:  # No cells
        return None

    counts_per_bc, plot_segments = sample_data.vdj_barcode_rank_plot_data()
    return _plot_segmented_barcode_rank(chart, counts_per_bc, plot_segments)


def plot_clonotype_table(chart, sample_properties, sample_data):
    if sample_data.vdj_clonotype_summary is None:
        return None

    clonotypes = sample_data.vdj_clonotype_summary.iloc[0:10]

    # This column used to be called 'cdr3s'; allow the webshim to work on older data
    cdr3_aa_col = "cdr3s_aa"
    if cdr3_aa_col not in clonotypes:
        cdr3_aa_col = "cdr3s"

    col_defs = collections.OrderedDict(
        [
            (
                "clonotype_id",
                {
                    "label": "Clonotype ID",
                    "format": "string",
                    "title": "Clonotype ID",
                    "style": "text-align: left",
                },
            ),
            (
                cdr3_aa_col,
                {
                    "label": "CDR3s",
                    "format": "string",
                    "title": "CDR3s in clonotype",
                    "style": "text-align: left",
                },
            ),
            (
                "frequency",
                {
                    "label": "Frequency",
                    "format": "integer",
                    "title": "Number of cells with clonotype",
                    "style": "text-align: right",
                },
            ),
            (
                "proportion",
                {
                    "label": "Proportion",
                    "format": "%0.4f",
                    "title": "Fraction of cell with clonotype",
                    "style": "text-align: right",
                },
            ),
        ]
    )

    cols = []
    for name, col_def in iteritems(col_defs):
        if name not in clonotypes:
            raise ValueError("Column not found in clonotype summary: %s" % name)
        cols.append(
            {"label": col_defs[name]["label"], "title": col_defs[name]["title"],}
        )

    rows = []
    for _, cl_row in clonotypes.iterrows():
        row = []
        for col_name, col_def in iteritems(col_defs):
            value = cl_row[col_name]
            formatted_value = format_value(value, col_def["format"])

            # Make the CDR3 list bit more readable
            formatted_value = formatted_value.replace(";", "; ")

            row.append(
                {
                    "v": tk_safe_json.json_sanitize(value),
                    "f": formatted_value,
                    "s": col_def["style"],
                }
            )
        rows.append(row)

    chart["table"].update({"rows": rows, "cols": cols})

    return chart


def plot_histogram_metric(chart, sample_properties, sample_data, **kwargs):
    """ Plot a HistogramMetric from the summary json """
    summary_data = sample_data.summary
    items = list(summary_data.get(kwargs["metric_name"], {}).items())
    if len(items) < 1:
        return None

    ordering = kwargs.get("order_by", shared_constants.HISTOGRAM_METRIC_DEFAULT_ORDERING)

    if ordering == shared_constants.HISTOGRAM_METRIC_ORDER_INTEGER_BIN:
        items.sort(key=lambda x: nullable_int_sort_key(x[0]))

    elif ordering == shared_constants.HISTOGRAM_METRIC_ORDER_DECREASING_FREQUENCY:
        items.sort(key=lambda x: nullable_int_sort_key(x[1], negate=True))

    elif ordering == shared_constants.HISTOGRAM_METRIC_ORDER_DECREASING_PROPORTION:
        items.sort(key=lambda x: -convert_to_float_gracefully(x[1]))

    x, y = zip(*items)
    chart["data"][0].update({"x": x, "y": y})

    return chart


def plot_barnyard_barcode_counts(chart, sample_properties, sample_data):
    analysis = sample_data.get_analysis(MultiGenomeAnalysis)
    if analysis is None:
        return None

    chart["data"] = []
    for label_info in ws_gex_constants.GEM_CALL_LABELS:
        name = label_info["label"]
        name = name.replace("genome0", analysis.result["genome0"])
        name = name.replace("genome1", analysis.result["genome1"])
        chart["data"].append(
            {
                "x": [],
                "y": [],
                "name": name,
                "mode": "markers",
                "marker": {"color": label_info["color"],},
            }
        )

    call_to_series = {v["key"]: i for i, v in enumerate(ws_gex_constants.GEM_CALL_LABELS)}
    for count0, count1, call in zip(
        analysis.result["count0"], analysis.result["count1"], analysis.result["call"]
    ):
        series = chart["data"][call_to_series[call]]
        series["x"].append(int(count0))
        series["y"].append(int(count1))

    chart["layout"]["xaxis"] = {
        "title": "%s UMI counts" % analysis.result["genome0"],
        "rangemode": "tozero",
        "autorange": True,
        "fixedrange": False,
    }
    chart["layout"]["yaxis"] = {
        "title": "%s UMI counts" % analysis.result["genome1"],
        "rangemode": "tozero",
        "autorange": True,
        "fixedrange": False,
    }

    return chart


def plot_preprocess(analyses):
    if analyses is None or len(analyses) == 0:
        return None

    sg_analyses = [an for an in analyses if isinstance(an, SingleGenomeAnalysis)]
    sg_analysis = sg_analyses[0] if len(sg_analyses) > 0 else None
    if sg_analysis is None or sg_analysis.is_zero_matrix():
        return None

    # Limit the number of K-means clusterings displayed to limit the HTML filesize
    new_clusterings = {}
    for key, clu in iteritems(sg_analysis.clusterings):
        if not (
            clu.clustering_type == cr_clustering.CLUSTER_TYPE_KMEANS
            and clu.num_clusters > ws_gex_constants.MAX_WEBSHIM_KMEANS_K
        ):
            new_clusterings[key] = clu
    sg_analysis.clusterings = new_clusterings

    return analyses


def load_sample_data(sample_properties, sample_data_paths):
    return SampleData(sample_properties, sample_data_paths, plot_preprocess)


def plot_tsne(chart, sample_properties, sample_data):
    """ Plot cells in t-SNE space, colored by clustering label """
    analysis = sample_data.get_analysis(SingleGenomeAnalysis)
    if analysis is None:
        return None

    args = [
        analysis.get_tsne().transformed_tsne_matrix,
        ws_gex_constants.TSNE_CLUSTER_DESCRIPTION,
        1,
        2,
    ]
    return clustering_plot_func(chart, sample_properties, sample_data, plot_dimensions, args)


def plot_dimensions(
    chart,
    transformed_matrix,
    description,
    pc1,
    pc2,
    clip=None,
    clustering=None,
    diff_expr=None,
    values=None,
    original_cluster_sizes=None,
):
    """ Plot cells in a 2-d space, colored by clustering label """
    assert (values is not None and clustering is None) or (
        clustering is not None and values is None and original_cluster_sizes is not None
    )

    if clustering is not None:
        values = clustering.clusters

    if original_cluster_sizes is None:
        value_freqs = np.bincount(values)[1:].tolist()
    else:
        # Cluster size array starts w/ cluster1 at index0; make it equivalent to np.bincount on arbitrary values
        value_freqs = [0] + original_cluster_sizes.tolist()

    n, m = transformed_matrix.shape
    if m < max(pc1, pc2):
        return None

    max_value = values.max()

    chart["data"] = [None] * max_value
    for value in np.unique(values):
        chart["data"][value - 1] = {
            "x": [],
            "y": [],
            "name": "%s %s - %s cells"
            % (
                description,
                format_value(value, "integer"),
                format_value(value_freqs[value], "integer"),
            ),
            "hoverinfo": "name",
            "mode": "markers",
            "type": "scattergl",
            "marker": {"size": 4,},
        }

    for i in range(n):
        r1 = shared_constants.DATA_VALUE_FORMAT % transformed_matrix[i, pc1 - 1]
        r2 = shared_constants.DATA_VALUE_FORMAT % transformed_matrix[i, pc2 - 1]
        value = values[i]

        series = chart["data"][value - 1]
        series["x"].append(r1)
        series["y"].append(r2)

    if clip is not None:
        xmin, xmax = np.percentile(transformed_matrix[:, pc1 - 1], clip)
        ymin, ymax = np.percentile(transformed_matrix[:, pc2 - 1], clip)
        chart["layout"]["xaxis"] = {
            "range": [xmin, xmax],
        }
        chart["layout"]["yaxis"] = {
            "range": [ymin, ymax],
        }

    chart["config"] = shared_constants.CHARTS_PLOTLY_MOVABLE_CONFIG
    return chart


def plot_dimensions_color(chart, transformed_matrix, values, description, vmin, vmax, pc1, pc2):
    _, m = transformed_matrix.shape
    if m < max(pc1, pc2):
        return None

    series = chart["data"][0]

    index_order = list(range(transformed_matrix.shape[0]))
    random.shuffle(index_order)
    for i in index_order:
        r1 = shared_constants.DATA_VALUE_FORMAT % transformed_matrix[i, pc1 - 1]
        r2 = shared_constants.DATA_VALUE_FORMAT % transformed_matrix[i, pc2 - 1]
        value = int(values[i])
        text = "%s: %s" % (description, format_value(value, "integer"))

        if value < vmin:
            value = vmin
        if value > vmax:
            value = vmax

        series["x"].append(r1)
        series["y"].append(r2)
        series["marker"]["color"].append(value)
        series["text"].append(text)

    return chart


def plot_tsne_totalcounts(chart, sample_properties, sample_data):
    """ Plot cells colored by total counts """
    analysis = sample_data.get_analysis(SingleGenomeAnalysis)
    if not analysis:
        return None

    reads_per_bc = analysis.matrix.get_counts_per_bc()
    vmin, vmax = np.percentile(reads_per_bc, ws_gex_constants.TSNE_TOTALCOUNTS_PRCT_CLIP)

    return plot_dimensions_color(
        chart,
        analysis.get_tsne().transformed_tsne_matrix,
        reads_per_bc,
        ws_gex_constants.TSNE_TOTALCOUNTS_DESCRIPTION,
        vmin,
        vmax,
        1,
        2,
    )


def _plot_differential_expression(
    chart, analysis, clustering=None, diff_expr=None, original_cluster_sizes=None
):
    n_clusters = clustering.clusters.max()

    # Get the union of top DE genes
    top_genes = set()

    # Limit the number of entries in the DE table
    n_genes = int(np.floor(float(ws_gex_constants.MAX_DE_TABLE_ENTRIES) / (n_clusters ** 2)))
    if n_genes < 1:
        n_genes = 1
    elif n_genes > ws_gex_constants.MAX_TOP_N_GENES:
        n_genes = ws_gex_constants.MAX_TOP_N_GENES

    cols = [
        {"type": "string", "label": "Gene ID"},
        {"type": "string", "label": "Gene name"},
    ]

    for i in range(n_clusters):
        # Filter genes by mean count and sort by log2 fold-change, descending
        means = diff_expr.data[:, 0 + 3 * i]
        log2fcs = diff_expr.data[:, 1 + 3 * i]

        keep_indices = np.flatnonzero(means >= ws_gex_constants.TOP_DE_GENES_MIN_MEAN)
        top_gene_indices = keep_indices[log2fcs[keep_indices].argsort()[::-1]][:n_genes]

        for j in top_gene_indices:
            top_genes.add(analysis.matrix.int_to_feature_id(j))

        cols.append(
            {
                "type": "number",
                "label": "L2FC",
                "title": "Log2 fold-change in cluster %d vs other cells" % (i + 1),
            }
        )
        cols.append(
            {
                "type": "number",
                "label": "p-value",
                "title": "Adjusted p-value of differential expression in cluster %d" % (i + 1),
            }
        )

    rows = []
    for gene_id in top_genes:
        i = analysis.matrix.feature_id_to_int(gene_id)
        gene_name = analysis.matrix.feature_id_to_name(gene_id)

        row = [gene_id, gene_name]
        for j in range(n_clusters):
            log2fc = diff_expr.data[i, 1 + (3 * j)]
            adj_p_value = diff_expr.data[i, 2 + (3 * j)]

            if log2fc <= 0 or adj_p_value >= ws_gex_constants.PVALUE_DEEMPHASIS_CUTOFF:
                style = "#DDD"
            else:
                style = "#000"

            row.append(
                {
                    "v": tk_safe_json.json_sanitize(log2fc),
                    "f": format_value(log2fc, "%.2f"),
                    "s": style,
                }
            )
            row.append(
                {
                    "v": tk_safe_json.json_sanitize(adj_p_value),
                    "f": format_value(adj_p_value, "%.0e"),
                    "s": style,
                }
            )

        rows.append(row)

    # Sort by log2fc, descending, in first cluster
    if n_clusters > 0:
        rows = sorted(rows, key=lambda row: row[2]["v"], reverse=True)

    chart["table"].update({"rows": rows, "cols": cols})
    return chart


def plot_differential_expression(chart, sample_properties, sample_data):
    sg_analysis = sample_data.get_analysis(SingleGenomeAnalysis)
    return clustering_plot_func(
        chart, sample_properties, sample_data, _plot_differential_expression, [sg_analysis]
    )


def clustering_plot_func(chart, sample_properties, sample_data, plot_func, args=[], kwargs={}):
    analysis = sample_data.get_analysis(SingleGenomeAnalysis)
    if analysis is None:
        return None

    new_charts = []
    for clustering_key, clustering in iteritems(analysis.clusterings):
        kwargs["clustering"] = clustering
        kwargs["original_cluster_sizes"] = sample_data.original_cluster_sizes[clustering_key]
        kwargs["diff_expr"] = analysis.differential_expression[clustering_key]
        new_chart = plot_func(copy.deepcopy(chart), *args, **kwargs)
        if new_chart is not None:
            new_chart["filters"] = {ws_gex_constants.CLUSTERS_FILTER_TITLE: clustering.description}
            new_charts.append(new_chart)
    return new_charts


def make_chart_filters(analyses):
    if analyses is None:
        return {}
    sg_analyses = [an for an in analyses if isinstance(an, SingleGenomeAnalysis)]
    if len(sg_analyses) == 0 or analyses[0] is None:
        return {}
    assert len(sg_analyses) == 1
    analysis = sg_analyses[0]

    filter_values = [
        x.description for x in cr_clustering.sort_clusterings(analysis.clusterings.values())
    ]

    return {
        ws_gex_constants.CLUSTERS_FILTER_TITLE: {
            "values": filter_values,
            "selected": filter_values[0],
        },
    }


def plot_subsampled_scatterplot_metric(chart, sample_properties, sample_data, **kwargs):
    """
    Modifies chart data entry to add traces for subsampled metrics with specified suffixes.
    Metrics must take the form <reference>_<subsample_type>_<subsample_depth>_<metric_suffix>, where metric suffix is specified as a kwarg

    KWARGS:
        metric_suffix (str): suffix for the subsampled metric given a metric of the form <reference>_<subsample_type>_<subsample_depth>_<metric_suffix>

    """
    summary_data = sample_data.summary or {}

    subsample_type = kwargs.get("subsample_type", "raw_rpc")

    # Just use ref_prefix if specified; otherwise try to construct it out of the set of references
    if "ref_prefix" in kwargs:
        ref_prefix = kwargs["ref_prefix"]
    elif "references" in kwargs:
        references = kwargs["references"]
        ref_prefix = "(" + "|".join(references) + ")_"
    else:
        ref_prefix = "(.+)_"

    # is the sample targeted, which lines to show
    is_targeted = kwargs.get("is_targeted", False)
    if kwargs.get("show_targeted_only", False):
        targeting_groups_to_plot = [cr_constants.ON_TARGET_SUBSAMPLE]
    else:
        targeting_groups_to_plot = [
            cr_constants.ON_TARGET_SUBSAMPLE,
            cr_constants.OFF_TARGET_SUBSAMPLE,
        ]

    # Regular expression to match <reference>_<subsample_type>_<subsample_depth>_<metric_suffix> pattern
    metric_pattern = r"^%s(%s)_([0-9]+)_%s_*(\w*)" % (
        ref_prefix,
        subsample_type,
        kwargs.get("metric_suffix", None),
    )

    metric_search_results = [re.search(metric_pattern, key) for key in summary_data.keys()]

    # Only return chart when multiple references are present for multi-genome only metrics
    if kwargs.get("multi_genome_only", False):
        references = {result.group(1) for result in metric_search_results if result is not None}
        if len(references) <= 2:
            return None

    # Find relevant metrics and extract data
    points = []

    for val, search_result in zip(summary_data.values(), metric_search_results):
        if search_result and not isinstance(val, float):
            val = float(val)
        # Only count corresponding keys and non-zero values
        if search_result and val > 0:
            genome = search_result.group(1)
            subsample_type = search_result.group(2).replace("_", " ")
            targeting_group = search_result.group(4)

            if kwargs.get("show_multi_genome_only", False) and (genome != "multi"):
                continue
            subsample_depth = int(search_result.group(3))

            if is_targeted and targeting_group not in targeting_groups_to_plot:
                continue
            elif is_targeted:
                subsample_type = "{}_{}".format(subsample_type, targeting_group)

            if genome == "multi":
                if is_targeted:
                    trace_label = cr_tgt_utils.reformat_targeted_label(targeting_group)
                else:
                    trace_label = ""
            else:
                if is_targeted:
                    trace_label = "{}_{}".format(
                        genome, cr_tgt_utils.reformat_targeted_label(targeting_group)
                    )
                else:
                    trace_label = genome

            points.append((subsample_depth, val, trace_label, genome, subsample_type,))

    # Don't produce chart if there's no data
    if len(points) == 0:
        return None

    # Sort extracted points by subsample depth (makes line plot connect properly)
    sorted_points = sorted(points, key=lambda x: (x[2], x[0]))

    # Build up a set of traces for <reference>_<subsample_type> pairs
    traces = {}

    for point in sorted_points:
        depth, value, trace_label, genome, subsample_type = point
        trace_key = (genome, subsample_type)

        if trace_key not in traces:
            # add (0,0)
            traces[trace_key] = {
                "x": [0, depth],
                "y": [0, value],
                "name": trace_label,
                "mode": "lines",
                "line": {"width": 3,},
            }
        else:
            traces[trace_key]["x"].append(depth)
            traces[trace_key]["y"].append(value)

    # Set extents of saturation line
    if (
        chart.get("layout")
        and chart["layout"].get("shapes")
        and chart["layout"]["shapes"][0]["type"] == "line"
    ):
        chart["layout"]["shapes"][0]["x1"] = max([max(trace["x"]) for trace in itervalues(traces)])

    # Add new data into chart
    chart["data"] = [v for _, v in sorted(traces.items())]
    return chart


def build_charts(sample_properties, chart_dicts, sample_data, all_prefixes={}, module=None):
    modules = [module, globals()] if module else [globals()]

    filters = make_chart_filters(sample_data.analyses)

    charts = []
    for chart_dict in chart_dicts:
        chart_dict = copy.deepcopy(chart_dict)
        function = chart_dict.pop("function")
        f = None
        for mod in modules:
            f = mod.get(function)
            if f is not None:
                break

        if function is not None and f is None:
            raise ValueError('Could not find webshim chart function "%s"' % function)

        kwargs = chart_dict.pop("kwargs", {})  # type: Dict[str, Any]

        kwargs_prefixes = chart_dict.pop("kwargs_prefixes", [])  # type: List[str]
        for prefix in kwargs_prefixes:
            kwargs[prefix] = all_prefixes[prefix]

        new_chart_obj = f(chart_dict, sample_properties, sample_data, **kwargs)
        if new_chart_obj is None:
            continue

        new_charts = new_chart_obj if isinstance(new_chart_obj, list) else [new_chart_obj]
        charts.extend(new_charts)

    return charts, filters


def filter_vdj_prefixes(all_prefixes, sample_properties):
    """ Only get subset of metric prefix values """
    chain_filter = sample_properties.chain_type
    if chain_filter is None:
        return all_prefixes

    # NOTE: Assumes chains (TRA, TRB, etc) are prefixed with the chain_type (TR or IG)
    result = {}
    for key, values in iteritems(all_prefixes):
        # Only filter any prefix that is a candidate for filtering
        # (i.e., contains some values prefixed by the selected chain type)
        if values is not None and any(v.startswith(chain_filter) for v in values):
            result[key] = [
                v
                for v in values
                if v.startswith(chain_filter) or v == rna_library.MULTI_REFS_PREFIX
            ]
        else:
            result[key] = values
    return result


def filter_vdj_alarms(all_alarms, sample_properties):
    """ Only get subset of metric alarms """
    chain_filter = sample_properties.chain_type
    if chain_filter is None:
        return all_alarms

    result = []
    for alarm in all_alarms:
        # No filters specified; don't filter
        if "filters" not in alarm:
            result.append(alarm)
            continue

        for f in alarm["filters"]:
            if f.get("chain_type") == chain_filter:
                result.append(alarm)

    return result


def get_constants_for_pipeline(pipeline, sample_properties):
    """ Get the appropriate metrics/alarms/charts for a pipeline """
    if pipeline == shared_constants.PIPELINE_VDJ:
        metrics, alarms, charts = (
            ws_vdj_constants.METRICS,
            ws_vdj_constants.METRIC_ALARMS,
            ws_vdj_constants.CHARTS,
        )

        metric_prefixes = filter_vdj_prefixes(
            vdj_report.VdjReporter().get_all_prefixes(), sample_properties
        )

        alarms = filter_vdj_alarms(alarms, sample_properties)

    else:
        metrics, alarms, charts = (
            ws_gex_constants.METRICS,
            ws_gex_constants.METRIC_ALARMS,
            ws_gex_constants.CHARTS,
        )

        metric_prefixes = cr_report.Reporter().get_all_prefixes()

    return metrics, alarms, charts, metric_prefixes


def get_custom_features(sample_data):
    """Infer the set of distinct custom feature types present in a dataset"""
    # Add the "custom feature type" prefix using the sample data
    analysis = sample_data.get_analysis(SingleGenomeAnalysis)

    # Adding a dummy value here suppresses extraneous web summary dashboards
    custom_features = ["dummy"]

    if (
        analysis
        and analysis.matrix.feature_ref.get_count_of_feature_type(rna_library.CUSTOM_LIBRARY_TYPE)
        > 0
    ):
        custom_features.append(rna_library.CUSTOM_LIBRARY_TYPE)
    return custom_features


def get_genomes(sample_data):
    """Infer the set of genomes present in a dataset"""

    analysis = sample_data.get_analysis(SingleGenomeAnalysis)
    if analysis:
        return [x for x in analysis.matrix.get_genomes() if x != ""]
    else:
        return ["dummy"]


def build_web_summary_json(sample_properties, sample_data, pipeline):
    """sample_properties - SampleProperties object
    sample_data - *SampleData class
    pipeline - string"""
    view = copy.deepcopy(sample_properties.__dict__)
    sample_properties_as_dict = copy.deepcopy(view)

    metrics, alarms, charts, all_prefixes = get_constants_for_pipeline(pipeline, sample_properties)

    all_prefixes["custom_features"] = get_custom_features(sample_data)

    tables, alarms = build_tables(
        sample_properties_as_dict, metrics, alarms, sample_data, all_prefixes=all_prefixes
    )
    if tables:
        view["tables"] = tables
    if alarms:
        view["alarms"] = alarms

    all_prefixes["references"] = list(set(all_prefixes["references"] + get_genomes(sample_data)))

    charts, filters = build_charts(
        sample_properties_as_dict, charts, sample_data, all_prefixes=all_prefixes
    )
    if charts:
        view["charts"] = charts
    if filters:
        view["filters"] = filters

    # Selected metrics that the web summary template needs
    info = build_info_dict(sample_properties, sample_data, pipeline)
    if info:
        view["info"] = info

    return view


def build_info_dict(sample_properties, sample_data, pipeline):
    """ Add miscellaneous metrics required by the web summary template """
    if sample_data.summary is None:
        return None
    info = {}

    info["chemistry_description"] = sample_data.summary.get("chemistry_description")

    info["references"] = []

    if pipeline in [shared_constants.PIPELINE_AGGR, shared_constants.PIPELINE_REANALYZE]:
        genomes = sample_properties.genomes
        if genomes is not None:
            info["references"].append(
                {
                    "type": cr_constants.REFERENCE_TYPE,
                    "name": cr_reference.get_ref_name_from_genomes(genomes),
                }
            )

    else:
        # Find all references in the summary
        reference_metric_prefixes = [
            cr_constants.REFERENCE_METRIC_PREFIX,
            vdj_constants.REFERENCE_METRIC_PREFIX,
        ]

        for prefix in reference_metric_prefixes:
            type_metric = "%s%s" % (prefix, cr_constants.REFERENCE_TYPE_KEY)

            if type_metric not in sample_data.summary:
                continue

            ref_type = sample_data.summary[type_metric]

            name_metric = "%s%s" % (prefix, cr_constants.REFERENCE_GENOMES_KEY)
            if name_metric not in sample_data.summary:
                raise ValueError(
                    "Reference metadata metric %s not found in metrics summary." % name_metric
                )
            ref_name = sample_data.summary.get(name_metric)

            if isinstance(ref_name, list):
                ref_name = cr_reference.get_ref_name_from_genomes(ref_name)

            info["references"].append({"type": ref_type, "name": ref_name})

    return info


def build_metrics_summary_csv(filename, sample_properties, sample_data, pipeline):
    metrics, alarms, _, all_prefixes = get_constants_for_pipeline(pipeline, sample_properties)

    sample_properties_as_dict = copy.deepcopy(sample_properties.__dict__)

    all_prefixes["custom_features"] = get_custom_features(sample_data)

    tables, _ = build_tables(
        sample_properties_as_dict, metrics, alarms, sample_data, all_prefixes=all_prefixes
    )
    if not tables:
        sys.stderr.write("No metrics tables were generated, skipping CSV generation.\n")
        return

    csv_metrics = collections.OrderedDict()
    for table in tables:
        if not table:
            continue
        for metric, _, value in table["rows"]:
            if isinstance(metric, dict):
                metric = metric["v"]
            if isinstance(value, dict):
                value = value["v"]
            if metric not in csv_metrics:
                csv_metrics[metric] = value

    with open(filename, "w") as f:
        writer = csv.writer(f, lineterminator="\n")
        writer.writerow(csv_metrics.keys())  # pylint: disable=dict-keys-not-iterating
        writer.writerow(csv_metrics.values())  # pylint: disable=dict-values-not-iterating


def nullable_int_sort_key(s, negate=False):
    """Return a sort key for s that will sort
    by int(s). Values of s that cannot be
    converted to an integer will be sorted
    to the begining."""

    try:
        if not negate:
            return (True, int(s))
        else:
            return (True, -int(s))
    except ValueError:
        return (False, 0)


def convert_to_float_gracefully(s):
    try:
        return float(s)
    except ValueError:
        return sys.float_info.max


def _make_jibes_color_map(tags):

    """
    Make an economist style color map for the websummary
    :param tags: list of tags
    :return: dict of {tag:color}
    """
    color_map = {}
    for i, tag in enumerate(sorted(tags)):
        color_map[ensure_str(tag)] = JIBES_COLOR_PALETTE[i]
    color_map[ensure_str(jibes.MULTIPLETS_FACTOR_NAME)] = JIBES_COLOR_PALETTE[12]
    color_map[ensure_str(jibes.BLANK_FACTOR_NAME)] = JIBES_COLOR_PALETTE[13]
    color_map[ensure_str(jibes.UNASSIGNED_FACTOR_NAME)] = JIBES_COLOR_PALETTE[14]
    return color_map


def make_jibes_histogram_plot(tag_names, counts, color_map):
    """
    :param tag names: labels for tags
     counts: the data to plot
     color_map: the colors to plot
    :return:
    jibes histogram data for websummary JSON
    """
    traces = []
    for i, tag in enumerate(tag_names):
        trace = {
            "x": counts[i],
            "type": "histogram",
            "opacity": 0.6,
            "name": ensure_str(tag),
            "marker": {"color": color_map[ensure_str(tag)],},
        }
        traces.append(trace)
    plot = {
        "data": traces,
        "layout": {"barmode": "overlay", "xaxis": {"title": "Log 10 (1 + Count)"}},
    }
    return plot


# the shared_resources implementation in python doesn't interop with Rust websummary at all
# need to just define a key to put the data under and use that
JIBES_SHARED_VECTORS_RESOURCE_KEY_PREFIX = "shared_jibes_data_vector"
RESOURCE_PREFIX = "_resources"


def make_jibes_biplot_histogram(assigner):
    """
    :param fitter: instance of JibesTagAssigner
    :return:
    dict containing jibes biplot and histogram data for websummary JSON
    """

    # pylint: disable=too-many-locals
    isinstance(assigner, JibesTagAssigner)
    fitter = assigner.fitter
    if fitter is None:
        return None
    assignments = assigner.jibes_assignments

    # we want roughly JIBES_BIPLOT_N_POINTS so find out how many rows to subsample
    n_rows_to_downsample = int(JIBES_BIPLOT_N_POINTS)
    # create a random mask for the rows with that many points
    downsample_indices = np.full(fitter.data.counts.shape[0], False)
    downsample_indices[:n_rows_to_downsample] = True
    np.random.shuffle(downsample_indices)

    # downsample the assignments
    assignments = assignments[downsample_indices]

    # downsample the data vectors
    original_vectors = [
        np.asarray(fitter.data.counts[downsample_indices, i]).ravel().tolist()
        for i in range(len(fitter.data.column_names))
    ]

    for vector in original_vectors:
        assert len(vector) == len(assignments)

    shared_resources = {}
    # this is sort of duplicating functionality that should be encapsulated by ResourceCollection in react_components.py
    # but we are sort of doing this in an ad-hoc way and passing it off to the Rust websummary code to
    # put it in the right place
    vectors = []
    for i, v in enumerate(original_vectors):
        store_key = "{}_{}".format(JIBES_SHARED_VECTORS_RESOURCE_KEY_PREFIX, i)
        use_key = "{}_{}".format(RESOURCE_PREFIX, store_key)
        shared_resources[store_key] = rc.array_to_float32_base64(v)
        vectors.append(use_key)

    vector_names = fitter.data.column_names
    color_map = _make_jibes_color_map(vector_names)
    histogram_data = make_jibes_histogram_plot(vector_names, vectors, color_map)

    # Clamp confidence values to avoid numeric issues
    assn_probs = assignments[jibes.ASSIGNMENT_PROB_COL_NAME].values
    ### Numeric issues values close to 1 by a small delta, sanity check we're close to
    # the expected range, then clamp the values to the QV min/max.
    over_1 = np.sum(assn_probs > 1.0001).sum() > 0
    under_0 = np.sum(assn_probs < -0.0001).sum() > 0
    if over_1 or under_0:
        # Should never ever happen
        raise Exception(
            "Normalized probabilities were outside of tolerance ranges.  This likely indicates a bug, please contact support@10xgenomics.com."
        )
    max_confidence = 1 - np.power(10, -(MAX_QV_VAL / 10.0))
    min_confidence = 1 - np.power(10, -(MIN_QV_VAL / 10.0))
    assn_probs = assn_probs.clip(min_confidence, max_confidence)
    phred_values = -10.0 * np.log10(1.0 - assn_probs)
    phred_values = phred_values.tolist()
    slider_values = [-10.0 * np.log10(1.0 - jibes.JIBES_MIN_CONFIDENCE)]

    biplot_data = rc.BiplotsWithThresholding(
        vectors,
        vector_names,
        True,
        False,
        phred_values,
        slider_values,
        "Assignment Phred Score Threshold",
        "Singlets",
        "Filtered Singlets",
        "% Filtered",
        [
            ensure_str(jibes.BLANK_FACTOR_NAME),
            ensure_str(jibes.MULTIPLETS_FACTOR_NAME),
            ensure_str(jibes.UNASSIGNED_FACTOR_NAME),
        ],
        [ensure_str(x) for x in assignments[jibes.ASSIGNMENT_COL_NAME].values.tolist()],
        color_map,
    )

    jibes_biplot_histogram = {
        "biplot": biplot_data,
        "histogram": histogram_data,
        "_resources": shared_resources,
    }

    return jibes_biplot_histogram



#
# Copyright (c) 2020 10X Genomics, Inc. All rights reserved.
#

from __future__ import absolute_import

import h5py
import numpy as np
import six

from bmkcc.cellranger.wrapped_tables import tables
import bmkcc.cellranger.h5_constants as h5_constants

if six.PY3:
    from html import unescape  # pylint: disable=no-name-in-module
    from typing import List, Optional, Union  # pylint: disable=unused-import
else:
    # Instance for unescaping, so we don't need to create one every time we decode.
    _HTML_PARSER = six.moves.html_parser.HTMLParser()

    def unescape(in_str):  # pylint: disable=function-redefined
        """Convert all named and numeric character references (e.g.
        `&gt;`, `&#62;`, `&#x3e;`) in the string `in_str` to the
        corresponding Unicode characters.
        """
        return _HTML_PARSER.unescape(in_str)


STR_DTYPE_CHAR = "S"
UNICODE_DTYPE_CHAR = "U"
STRING_DTYPE_CHARS = (STR_DTYPE_CHAR, UNICODE_DTYPE_CHAR)


def is_hdf5(filename, throw_exception=False):
    """
    A wrapper around h5py.is_hdf5, optionally can throw an exception.

    Args:
        filename: The name of the file to test
        throw_exception: Should we raise an error if not?

    Returns:
        bool as to whether the file is valid
    """
    valid = h5py.is_hdf5(filename)
    if not valid and throw_exception:
        raise IOError("File: {} is not a valid HDF5 file.".format(filename))
    return valid


def write_h5(filename, data, append=False):
    filemode = "w"
    if append:
        filemode = "a"
    with h5py.File(filename, filemode) as f:
        for key, value in six.iteritems(data):
            f[key] = value


def get_h5_filetype(filename):
    with tables.open_file(filename, mode="r") as f:
        try:
            filetype = six.ensure_str(f.get_node_attr("/", h5_constants.H5_FILETYPE_KEY))
        except AttributeError:
            filetype = None  # older files lack this key
    return filetype


def save_array_h5(filename, name, arr):
    """ Save an array to the root of an h5 file """
    with tables.open_file(filename, "w") as f:
        f.create_carray(f.root, name, obj=arr)


def load_array_h5(filename, name):
    """ Load an array from the root of an h5 file """
    with tables.open_file(filename, "r") as f:
        return getattr(f.root, name).read()


def create_hdf5_string_dataset(group, name, data, **kwargs):
    """Create a dataset of strings under an HDF5 (h5py) group.

    Strings are stored as fixed-length 7-bit ASCII with XML-encoding
    for characters outside of 7-bit ASCII. This is inspired by the
    choice made for the Loom spec:
    https://github.com/linnarsson-lab/loompy/blob/master/doc/format/index.rst

    Args:
        group (h5py.Node): Parent group.
        name (str): Dataset name.
        data (list of str): Data to store. Both None and [] are serialized to an empty dataset.
                            Both elements that are empty strings and elements that are None are
                            serialized to empty strings.
    """

    if data is None or hasattr(data, "__len__") and len(data) == 0:
        dtype = "%s1" % STR_DTYPE_CHAR
        group.create_dataset(name, dtype=dtype)
        return

    assert (isinstance(data, np.ndarray) and data.dtype.char == STR_DTYPE_CHAR) or (
        isinstance(data, list)
        and all(x is None or isinstance(x, (bytes, six.text_type)) for x in data)
    )

    # Convert Nones to empty strings and use XML encoding
    data = [encode_ascii_xml(x) if x else b"" for x in data]

    fixed_len = max(len(x) for x in data)

    # h5py doesn't support strings with zero-length-dtype
    if fixed_len == 0:
        fixed_len = 1
    dtype = "%s%d" % (STR_DTYPE_CHAR, fixed_len)

    group.create_dataset(name, data=data, dtype=dtype, **kwargs)


def decode_ascii_xml(x):
    # type: (Union[str, bytes]) -> str
    """
    Decode a string from 7-bit ASCII + XML into unicode.
    """
    if isinstance(x, six.text_type):
        return x
    elif isinstance(x, six.binary_type):
        return six.ensure_text(unescape(six.ensure_str(x)))
    else:
        raise ValueError("Expected string type, got type %s" % str(type(x)))


def decode_ascii_xml_array(data):
    """Decode an array-like container of strings from 7-bit ASCII + XML
    into unicode.
    """
    if isinstance(data, np.ndarray) and data.dtype.char == UNICODE_DTYPE_CHAR:
        return data

    unicode_data = [decode_ascii_xml(x) for x in data]

    fixed_len = max(len(s) for s in unicode_data)
    # 0 length string type is Ok
    dtype = "%s%d" % (UNICODE_DTYPE_CHAR, fixed_len)

    # note: python3 would require no.fromiter
    return np.array(unicode_data, dtype=dtype)


def encode_ascii_xml(x):
    """Encode a string as fixed-length 7-bit ASCII with XML-encoding
    for characters outside of 7-bit ASCII.

    Respect python2 and python3, either unicode or binary.
    """
    if isinstance(x, six.text_type):
        return x.encode("ascii", "xmlcharrefreplace")
    elif isinstance(x, six.binary_type):
        return x
    else:
        raise ValueError("Expected string type, got type %s" % str(type(x)))


def encode_ascii_xml_array(data):
    """Encode an array-like container of strings as fixed-length 7-bit ASCII
    with XML-encoding for characters outside of 7-bit ASCII.
    """
    if (
        isinstance(data, np.ndarray)
        and data.dtype.char == STR_DTYPE_CHAR
        and data.dtype.itemsize > 0
    ):
        return data

    convert = lambda s: encode_ascii_xml(s) if s is not None else b""
    ascii_data = [convert(x) for x in data]

    fixed_len = max(len(s) for s in ascii_data)
    fixed_len = max(1, fixed_len)
    dtype = "%s%d" % (STR_DTYPE_CHAR, fixed_len)

    # note: python3 would require np.fromiter
    return np.array(ascii_data, dtype=dtype)


# Global variables used in memoization below
_mlastConverted = None
_mlastValue = None  # type: Optional[str]


def read_hdf5_string_dataset(dataset, memoize=False):
    # type: (h5py.Dataset, bool) -> List[str]
    """Read a dataset of strings from HDF5 (h5py).

    Args:
        dataset (h5py.Dataset): Data to read.
        memoize (bool): Whether to use memoization to make string conversion more efficient.

    Returns:
        list[unicode]: Strings in the dataset.
    """

    # h5py doesn't support loading an empty dataset
    if dataset.shape is None:
        return []

    data = dataset[:]
    decode = decode_ascii_xml
    if memoize:

        def decoder(x):
            # type: (Union[str, bytes]) -> str
            # pylint: disable=global-statement
            global _mlastConverted
            global _mlastValue
            if x != _mlastConverted:
                _mlastValue = decode_ascii_xml(x)
                _mlastConverted = x
            assert _mlastValue is not None
            return _mlastValue

        decode = decoder
    return [decode(x) for x in data]


def set_hdf5_attr(dataset, name, value):
    """Set an attribute of an HDF5 dataset/group.

    Strings are stored as fixed-length 7-bit ASCII with XML-encoding
    for characters outside of 7-bit ASCII. This is inspired by the
    choice made for the Loom spec:
    https://github.com/linnarsson-lab/loompy/blob/master/doc/format/index.rst
    """

    name = encode_ascii_xml(name)

    if isinstance(value, (six.text_type, six.binary_type)):
        value = encode_ascii_xml(value)
    elif isinstance(value, np.ndarray) and value.dtype.char in STRING_DTYPE_CHARS:
        value = encode_ascii_xml_array(value)
    elif isinstance(value, (list, tuple)) and isinstance(
        value[0], (six.text_type, six.binary_type)
    ):
        value = encode_ascii_xml_array(value)

    dataset.attrs[name] = value


def combine_h5s_into_one(outfile, files_to_combine):
    # type: (str, list) -> None
    assert isinstance(files_to_combine, list)
    with tables.open_file(six.ensure_binary(outfile), "a") as out:
        for fname in files_to_combine:
            with tables.open_file(six.ensure_binary(fname), "r") as data:
                data.copy_children(data.root, out.root, recursive=True)



#!/usr/bin/env python
#
# Copyright (c) 2019 10X Genomics, Inc. All rights reserved.
#

""" Commonly-used constants which have not yet been organized. """

######################################################
# DO NOT add new items to this file.
#
# - If a constant is only used from a single module, put it in that module.
# - If a constant is only used in association with a particular module, put it
#   in that module.
# - If a constant is used in a bunch of places, create a module with a more
#   specifically descriptive name to put those constants.
######################################################

from __future__ import absolute_import

import os
from collections import namedtuple

Interval = namedtuple("Interval", ["chrom", "start", "end", "length", "strand"])
Transcript = namedtuple("Transcript", ["gene", "length", "gc_content", "intervals"])
Gene = namedtuple("Gene", ["id", "name", "length", "gc_content", "intervals"])

BARCODE_WHITELIST_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "barcodes")
BARCODE_WHITELIST_TRANSLATE_PATH = os.path.join(BARCODE_WHITELIST_PATH, "translation")

NO_INPUT_FASTQS_MESSAGE = "No input FASTQs were found with the requested sample indices."

# Product types defs
SPATIAL_PRODUCT_TYPE = "sp"
SINGLE_CELL_PRODUCT_TYPE = "sc"

# Bam tags
# NUM_HITS_TAG = "NH"
MULTIMAPPER_TAG = "MM"
ANTISENSE_TAG = "AN"
RAW_BARCODE_TAG = "CR"
PROCESSED_BARCODE_TAG = "CB"
RAW_BARCODE_QUAL_TAG = "CY"
RAW_UMI_TAG = "UR"
PROCESSED_UMI_TAG = "UB"
UMI_QUAL_TAG = "UY"
TRANSCRIPTS_TAG = "TX"
GENE_IDS_TAG = "GX"
# GENE_NAMES_TAG = "GN"
MAPPING_REGION_TAG = "RE"
RAW_FEATURE_BARCODE_TAG = "fr"
PROCESSED_FEATURE_BARCODE_TAG = "fb"
FEATURE_BARCODE_QUAL_TAG = "fq"
FEATURE_IDS_TAG = "fx"
LIBRARY_INDEX_TAG = "li"
EXTRA_FLAGS_TAG = "xf"
# MATE_RESCUE_TAG = "MR"


# Cell Ranger read types:
#     all: all reads
#     mapped: defined per region
#     conf_mapped: defined per region
#     conf_mapped_barcoded: read must be conf_mapped and have valid BC if BCs are used
#     conf_mapped_deduped: read must be conf_mapped_barcoded, non-duplicate and have valid UMI if UMIs are used
# ALL_READ_TYPE = "all"
# MAPPED_READ_TYPE = "mapped"
CONF_MAPPED_READ_TYPE = "conf_mapped"
CONF_MAPPED_BC_READ_TYPE = "conf_mapped_barcoded"
CONF_MAPPED_DEDUPED_READ_TYPE = "conf_mapped_deduped_barcoded"

# Molecule types:
#     insert: a distinct sequenced read.
#     fragment: a called siPCR fragment molecule; a distinct (BC, UMI, gene, pos, strand) tuple
#     cDNA: a called cDNA molecule; a distinct (BC, UMI, gene) tuple
#     cDNA_candidate: a candidate cDNA molecule; a distinct (BC, UMI) tuple
INSERT_MOLECULE_TYPE = "insert"
FRAGMENT_MOLECULE_TYPE = "fragment"
CDNA_MOLECULE_TYPE = "cdna"
# CDNA_MOLECULE_CANDIDATE_TYPE = "cdna_candidate"

# Barcode types:
#    all_bcs: all barcodes
#    filtered_bcs: cell-containing barcodes
ALL_BARCODES = "all_bcs"
FILTERED_BARCODES = "filtered_bcs"

# Regions:
#     transcriptome:
#       mapped: read must be mapped to sense strand of at least one transcript
#       conf_mapped: read must be confidently mapped along the sense strand to a unique gene
#     genome:
#       mapped: read must be mapped to at least one chromosome
#       conf_mapped: read must be confidently mapped to at least one chromosome
#     exonic:
#       mapped: read must be mapped to at least one exonic region
#       conf_mapped: read must be confidently mapped to at least one exonic region
#     intergenic:
#       mapped: read must be mapped to at least one intergenic region
#       conf_mapped: read must be confidently mapped to at least one intergenic region
#     intronic:
#       mapped: read must be mapped to at least one intronic region
#       conf_mapped: read must be confidently mapped to at least one intronic region
TRANSCRIPTOME_REGION = "transcriptome"
# GENOME_REGION = "genome"
EXONIC_REGION = "exonic"
INTERGENIC_REGION = "intergenic"
INTRONIC_REGION = "intronic"

# Duplicate types:
#     cdna_pcr_uncorrected: determined by gene ID, barcode, strand, uncorrected umi
#     cdna_pcr: determined by gene ID, barcode, strand, corrected umi
#     si_pcr: determined by gene ID, pos, barcode, strand, corrected umi
# CDNA_PCR_UNCORRECTED_DUPE_TYPE = "cdna_pcr_uncorrected"
CDNA_PCR_DUPE_TYPE = "cdna_pcr"
# SI_PCR_DUPE_TYPE = "si_pcr"

ON_TARGET_SUBSAMPLE = "ontarget"
OFF_TARGET_SUBSAMPLE = "offtarget"

FORWARD_STRAND = b"+"
REVERSE_STRAND = b"-"
STRANDS = [FORWARD_STRAND, REVERSE_STRAND]

THREE_PRIME = "three_prime"
FIVE_PRIME = "five_prime"

INSERT_SIZE_CUTOFFS = list(range(0, 1550, 50))
MIN_COUNTS_PER_GENE = 1
DEFAULT_RECOVERED_CELLS_PER_GEM_GROUP = 3000

TOP_N = 5
HOMOPOLYMER_LENGTH = 15

H5_BC_SEQUENCE_COL = "bc_sequence"

CELLRANGER_VERSION_KEY = "cellranger_version"

REFERENCE_METADATA_FILE = "reference.json"
REFERENCE_STAR_PATH = "star"
REFERENCE_FASTA_PATH = "fasta/genome.fa"
REFERENCE_GENES_GTF_PATH = "genes/genes.gtf"
REFERENCE_GENOMES_KEY = "genomes"
REFERENCE_MEM_GB_KEY = "mem_gb"
# REFERENCE_NUM_THREADS_KEY = "threads"

# Ref metadata keys used by GEX and VDJ
REFERENCE_FASTA_HASH_KEY = "fasta_hash"
REFERENCE_GTF_HASH_KEY = "gtf_hash"
REFERENCE_INPUT_FASTA_KEY = "input_fasta_files"
REFERENCE_INPUT_GTF_KEY = "input_gtf_files"
REFERENCE_MKREF_VERSION_KEY = "mkref_version"
REFERENCE_VERSION_KEY = "version"
REFERENCE_TYPE_KEY = "type"
REFERENCE_TYPE = "Transcriptome"
REFERENCE_METRIC_PREFIX = "reference_"

BAM_CHUNK_SIZE_GB = 0.5
MAX_BAM_CHUNKS = 256

# NOTE: MEM_GB_PER_THREAD is only used for a few stages where we've encountered memory oversubscription issues
# on clusters without memory reservations. As of 3/15/2017, it's used by:
# - RUN_PCA
# - RUN_DIFFERENTIAL_EXPRESSION
# - RUN_GRAPH_CLUSTERING
MEM_GB_PER_THREAD = 8

COUNT_GENES_MAX_MEM_GB = 64
NUM_MOLECULE_INFO_ENTRIES_PER_CHUNK = 40000000

STAR_DEFAULT_HIGH_CONF_MAPQ = 255

NORM_MODE_NONE = "none"

AGG_ID_FIELD = "library_id"
AGG_H5_FIELD = "molecule_h5"
AGG_BATCH_FIELD = "batch"
AGG_DESCRIPTION_FIELD = "description"

# Spatial AGGR specific
AGG_CLOUPE_FIELD = "cloupe_file"
AGG_SPATIAL_FIELD = "spatial_folder"
AGG_TISSUE_POSITION_FIELD = "tissue_position"
AGG_SCALE_FACTORS_FIELD = "scale_factors"
AGG_HIRES_IMAGES_FIELD = "hires_images"
AGG_LOWRES_IMAGES_FIELD = "lowres_images"

# Namedtuples definitions
agg_fields = namedtuple("agg_fields", ["type", "required"])
aggr_files = namedtuple("aggr_files", ["path", "required", "default_location"])

# Definition of the sc aggr files
SC_AGGR_FILES = {
    AGG_H5_FIELD: aggr_files(path="molecule_info.h5", required=True, default_location="")
}

# Definition of spatial aggr files
SPATIAL_AGGR_FILES = {
    AGG_H5_FIELD: aggr_files(path="molecule_info.h5", required=True, default_location=""),
    AGG_CLOUPE_FIELD: aggr_files(path="cloupe.cloupe", required=True, default_location=""),
    AGG_TISSUE_POSITION_FIELD: aggr_files(
        path="tissue_positions_list.csv", required=True, default_location="spatial"
    ),
    AGG_SCALE_FACTORS_FIELD: aggr_files(
        path="scalefactors_json.json", required=True, default_location="spatial"
    ),
    AGG_HIRES_IMAGES_FIELD: aggr_files(
        path="tissue_hires_image.png", required=True, default_location="spatial"
    ),
    AGG_LOWRES_IMAGES_FIELD: aggr_files(
        path="tissue_lowres_image.png", required=True, default_location="spatial"
    ),
}

MAX_INSERT_SIZE = 1000

# Chemistry detection
DETECT_CHEMISTRY_INITIAL_READS = 100000

# PIPELINE NAMES
PIPELINE_VDJ = "vdj"
PIPELINE_COUNT = "count"

# Barcode for non-whitelisted barcodes in the per barcode metrics CSV
NO_BARCODE = b"NO_BARCODE"

# constants for dark_images mro paramter
DARK_IMAGES_NONE = 0
DARK_IMAGES_CHANNELS = 1
DARK_IMAGES_COLORIZED = 2

# Spatial, tissue color
TISSUE_COLOR = "#E84B50"  # a color as understood by matplotlib, see https://matplotlib.org/2.0.2/api/colors_api.html

# alpha blending to apply to spots under tissue; used in both, the automatic and the manual alignment paths.
TISSUE_SPOTS_ALPHA = 0.50

# alpha blending to apply to tissue area outside of spots; used only in the automatic alignment.
# Has to be smaller than TISSUE_SPOTS_ALPHA.
TISSUE_NOSPOTS_ALPHA = 0.25

# color for the bounding box drawn around the tissue detection area
# this is an 8-bit RGB triple
TISSUE_BBOX_COLOR = (29, 67, 122)

# the color for the ring traced around fiducial spots on the QC image
FIDUCIAL_SPOT_COLOR = "#1D437A"
FILTER_LIST = [
    "None",
    "Non-Targeting",
    "Ignore",
]  # List of targets that are considered filter-able ie



#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#

"""
Functions for fast, memory-efficient operations on masked sparse matrices.
"""

from __future__ import absolute_import

import numpy as np
from six.moves import xrange as range


def sum_masked(matrix, row_mask, col_mask, axis):
    """Sum a masked sparse matrix along an axis.
    Args:
      matrix (scipy.sparse.csc_matrix): Matrix
      axis (int): Axis to sum away; None for total sum.
      row_mask (np.array): Boolean mask for rows
      col_mask (np.array): Boolean mask for columns
    Returns:
      np.array
    """
    assert matrix.getformat() == "csc"
    assert len(row_mask) == matrix.shape[0]
    assert len(col_mask) == matrix.shape[1]
    assert row_mask.dtype == "bool"
    assert col_mask.dtype == "bool"

    if axis == 0:
        return matrix.transpose(copy=False).dot(row_mask)[col_mask]

    elif axis == 1 or axis is None:
        row_sums = matrix.dot(col_mask)[row_mask]

        if axis == 1:
            return row_sums
        assert axis is None
        return np.sum(row_sums)

    else:
        raise ValueError("axis must be 0, 1, or None")


def count_ge_masked(m, row_mask, col_mask, threshold, axis, chunk_size_mb=64):
    """Count values greater than or equal to a threshold in a masked sparse matrix.

    Args:
      m (scipy.sparse.csc_matrix): Matrix
      axis (int): Axis to sum away; None for total count.
      threshold (int): Only count elements exceeding this threshold
      row_mask (np.array): Boolean mask for rows
      col_mask (np.array): Boolean mask for columns
      chunk_size_mb (int): Operate on chunks of this size; don't copy the entire matrix at once.
    Returns:
      np.array
    """
    assert chunk_size_mb > 0
    assert m.getformat() == "csc"
    assert len(row_mask) == m.shape[0]
    assert len(col_mask) == m.shape[1]
    assert row_mask.dtype == "bool"
    assert col_mask.dtype == "bool"

    # Operate on chunks of bounded size
    ind_bytes = m.indices.dtype.itemsize
    data_bytes = m.data.dtype.itemsize
    bytes_per_nz_elem = ind_bytes + data_bytes

    nz_elem_per_col = np.diff(m.indptr)
    bytes_per_col = nz_elem_per_col * bytes_per_nz_elem
    chunk_size_bytes = chunk_size_mb * (1024 * 1024)

    # Split the matrix into approxiamtely equal-RAM barcode chunks
    chunk_starts = [0]
    chunk_size = 0
    for i in range(len(bytes_per_col)):
        chunk_size += bytes_per_col[i]
        if chunk_size >= chunk_size_bytes:
            chunk_starts.append(i)
            chunk_size = 0

    results = []

    for chunk_idx, chunk_start in enumerate(chunk_starts):
        chunk_end = (
            chunk_starts[1 + chunk_idx] if chunk_idx < (len(chunk_starts) - 1) else m.shape[1]
        )
        chunk = slice(chunk_start, chunk_end)
        submatrix = (m[:, chunk] >= threshold).astype("uint64")
        if axis == 0:
            s = submatrix.transpose(copy=False).dot(row_mask)
        elif axis == 1 or axis is None:
            s = submatrix.dot(col_mask[chunk])
        results.append(s)

    if axis == 0:
        return np.concatenate(results)[col_mask]
    elif axis == 1:
        return sum(results)[row_mask]
    elif axis is None:
        return np.sum(sum(results)[row_mask])
    else:
        raise ValueError("Axis must be 0, 1, or None")



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#

from __future__ import absolute_import, division

import bmkcc.cellranger.analysis.io as analysis_io
import bmkcc.cellranger.analysis.constants as analysis_constants
import bmkcc.cellranger.io as cr_io

import collections
import numpy as np
import os
import tsne as tsne_bh

TSNE = collections.namedtuple(
    "TSNE",
    [
        "transformed_tsne_matrix",
        "name",  # Human readable form
        "key",  # Machine queryable form, must be unique
    ],
)


def run_tsne(
    transformed_pca_matrix,
    name="TSNE",
    key="TSNE",
    tsne_dims=None,
    input_pcs=None,
    perplexity=None,
    theta=None,
    max_iter=None,
    stop_lying_iter=None,
    mom_switch_iter=None,
    copy_data=False,
    random_state=None,
):

    if tsne_dims is None:
        tsne_dims = analysis_constants.TSNE_N_COMPONENTS

    if perplexity is None:
        perplexity = analysis_constants.TSNE_DEFAULT_PERPLEXITY

    if theta is None:
        theta = analysis_constants.TSNE_THETA

    if random_state is None:
        random_state = analysis_constants.RANDOM_STATE

    if max_iter is None:
        max_iter = analysis_constants.TSNE_MAX_ITER

    if stop_lying_iter is None:
        stop_lying_iter = analysis_constants.TSNE_STOP_LYING_ITER

    if mom_switch_iter is None:
        mom_switch_iter = analysis_constants.TSNE_MOM_SWITCH_ITER

    if input_pcs is not None:
        transformed_pca_matrix = transformed_pca_matrix[:, :input_pcs]

    # Make sure perplexity satisfies 'tsne' requirements
    N = transformed_pca_matrix.shape[0]
    perplexity = min(perplexity, max(1, -1 + float((N - 1)) / 3))

    # At sufficiently low cell count, we cannot satisfy perplexity >= 1 and this condition below,
    # so the projection is defined as all zeros
    if N - 1 < 3 * perplexity:
        transformed_tsne_matrix = np.zeros((N, tsne_dims))
    else:
        transformed_tsne_matrix = tsne_bh.bh_sne(
            transformed_pca_matrix,
            d=tsne_dims,
            theta=theta,
            perplexity=perplexity,
            max_iter=max_iter,
            stop_lying_iter=stop_lying_iter,
            mom_switch_iter=mom_switch_iter,
            copy_data=copy_data,
            random_state=np.random.RandomState(random_state),
        )

    return TSNE(transformed_tsne_matrix, name=name, key=key)


def save_tsne_csv(tsne, barcodes, base_dir):
    """Save a TSNE object to CSV"""
    # Preserve backward compatibility with pre-3.0 CSV files
    #   where the CSV directory was named "2_components" and the HDF5 dataset was named "_2"
    key = tsne.key + "_components"

    tsne_dir = os.path.join(base_dir, key)
    cr_io.makedirs(tsne_dir, allow_existing=True)

    matrix_fn = os.path.join(tsne_dir, "projection.csv")
    n_tsne_components = tsne.transformed_tsne_matrix.shape[1]
    matrix_header = ["Barcode"] + ["TSNE-%d" % (i + 1) for i in range(n_tsne_components)]
    analysis_io.save_matrix_csv(matrix_fn, tsne.transformed_tsne_matrix, matrix_header, barcodes)


def save_tsne_h5(tsne, fname):
    """Save a TSNE object to HDF5"""
    t_map = {tsne.key: tsne}
    analysis_io.save_dimension_reduction_h5(t_map, fname, analysis_constants.ANALYSIS_H5_TSNE_GROUP)




from __future__ import division

import numpy as np
cimport numpy as np

cpdef np.int32_t bisect_left(const np.int32_t[:] a, x, np.ndarray keys):
    cdef int lo = 0, hi = a.shape[0], i
    while hi - lo > 1:
        i = (lo + hi) // 2  # cannot be <1
        y = keys[a[i]]
        if keys[a[i-1]] < x and x <= y:
            return a[i]
        elif y < x:
            lo = i
        elif hi == i + 1:
            hi = i
        else:
            hi = i + 1
    return a[lo]
